from typing import *
from collections import namedtuple
import numpy as np

from .gate_type import MatrixType
import QuICT.ops.linalg.cpu_calculator as CPUCalculator

from QuICT.tools.exception.core import TypeError


def get_gates_order_by_depth(gates: list) -> list:
    """ Order the gates of circuit by its depth layer

    Returns:
        List[List[BasicGate]]: The list of gates which at same layers in circuit.
    """
    gate_by_depth = [[gates[0]]]          # List[list], gates for each depth level.
    # List[set], gates' qubits for each depth level.
    gate_args_by_depth = [set(gates[0].cargs + gates[0].targs)]
    for gate in gates[1:]:
        gate_arg = set(gate.cargs + gate.targs)
        for i in range(len(gate_args_by_depth) - 1, -1, -1):
            if gate_arg & gate_args_by_depth[i]:
                if i == len(gate_args_by_depth) - 1:
                    gate_by_depth.append([gate])
                    gate_args_by_depth.append(gate_arg)
                else:
                    gate_by_depth[i + 1].append(gate)
                    gate_args_by_depth[i + 1] = gate_arg | gate_args_by_depth[i + 1]
                break
            else:
                if i == 0:
                    gate_by_depth[i].append(gate)
                    gate_args_by_depth[i] = gate_arg | gate_args_by_depth[i]

    return gate_by_depth


class MatrixGroup:
    def __init__(self, matrix, args, blocked_args: set = set([])):
        self.args = set(args)
        self.value = [(matrix, args)]
        self.block_args = blocked_args

    def append(self, matrix, args):
        self.args = self.args | set(args)
        self.value.append((matrix, args))

    def intersect(self, args: list):
        intersect_qubits = list(set(args) & self.args)
        blocked_qubits = list(set(args) & self.block_args)

        return intersect_qubits, blocked_qubits


class CircuitMatrix:
    def __init__(self, device: str = "CPU"):
        self._device = device
        if device == "CPU":
            self._computer = CPUCalculator
            self._array_helper = np
        else:
            import cupy as cp
            import QuICT.ops.linalg.gpu_calculator as GPUCalculator

            self._computer = GPUCalculator
            self._array_helper = cp

    def get_unitary_matrix(self, gates: list, qubits_num: int, mini_arg: int = 0) -> np.ndarray:
        # Order gates by depth
        gates_order_by_depth = get_gates_order_by_depth(gates)

        # Get MatrixGroups by its qubit args and depth
        matrix_groups = [[]]          # List[List[MatrixGroup]]
        for layer_gates in gates_order_by_depth:
            for gate in layer_gates:
                if gate.controls + gate.targets >= 3:
                    raise TypeError(
                        "CircuitMatrix.get_unitary_matrix.gates", "1 or 2-qubits gates", gate.controls + gate.targets
                    )

                if gate.matrix_type == MatrixType.special:
                    continue

                args = gate.cargs + gate.targs
                matrix = gate.matrix if self._device == "CPU" else self._array_helper.array(gate.matrix)
                if len(args) == 2 and args[0] > args[1]:
                    args.sort()
                    matrix = self._computer.MatrixPermutation(matrix, self._array_helper.array([1, 0]))

                is_intersect, is_blocked_layer = self._find_related_MatrixGroup(matrix_groups, args)
                if not is_intersect:
                    new_mg = MatrixGroup(matrix, args)
                    matrix_groups[is_blocked_layer].append(new_mg)
                else:
                    if len(is_intersect) == 2:
                        related_mg0, related_mg1 = is_intersect[0], is_intersect[1]
                        if related_mg0.position == related_mg1.position:
                            matrix_groups[related_mg0.layer][related_mg0.position].append(matrix, args)
                        else:
                            blocked_args = matrix_groups[related_mg0.layer][related_mg0.position].args | \
                                matrix_groups[related_mg1.layer][related_mg1.position].args
                            new_mg = MatrixGroup(matrix, args, blocked_args)
                            if related_mg0.layer == len(matrix_groups) - 1:
                                matrix_groups.append([new_mg])
                            else:
                                matrix_groups[related_mg0.layer + 1].append(new_mg)
                    else:
                        intersect_info = is_intersect[0]
                        matrix_groups[intersect_info.layer][intersect_info.position].append(matrix, args)

        # Combined the matries in each MatrixGroup
        combined_matries = []
        for layer in matrix_groups:
            for mg in layer:
                combined_matries.append(self._combined_gates(mg.value))

        # Combined all matries from the combined MatrixGroup
        circuit_matrix, circuit_matrix_args = self._combined_gates(combined_matries)
        # Permutation the circuit matrix with currect qubits' order
        args_baseline = list(range(mini_arg, qubits_num, 1))
        if circuit_matrix_args != args_baseline:
            circuit_matrix = self._tensor_unitary(circuit_matrix, circuit_matrix_args, args_baseline)

        return circuit_matrix

    def _find_related_MatrixGroup(self, matrix_groups: list, args: list):
        intersect_info = namedtuple('IntersectInfo', ['arg', 'layer', 'position'])
        is_blocked = 0
        is_intersect = []
        for lidx in range(len(matrix_groups) - 1, -1, -1):
            layer = matrix_groups[lidx]
            for midx, mg in enumerate(layer):
                iq, bq = mg.intersect(args)
                if len(iq) > 0:
                    for idx in iq:
                        inter_info = intersect_info(idx, lidx, midx)
                        is_intersect.append(inter_info)

                if len(bq) > 0:
                    is_blocked = lidx

            if len(is_intersect) > 0:
                return is_intersect, False

            if is_blocked:
                break

        return is_intersect, is_blocked

    def merge_gates(self, u1, u1_args, u2, u2_args):
        u1_args_set, u2_args_set = set(u1_args), set(u2_args)
        insection_args = u1_args_set & u2_args_set
        if len(insection_args) == 0:
            return self._computer.tensor(u1, u2), u1_args + u2_args

        if u1_args_set == u2_args_set:
            args_idx = [u1_args.index(u2_arg) for u2_arg in u2_args]
            u2 = self._computer.MatrixPermutation(
                u2,
                self._array_helper.array(args_idx)
            )

        union_args = u1_args + [i for i in u2_args if i not in u1_args] if len(u1_args) >= len(u2_args) else \
            u2_args + [i for i in u1_args if i not in u2_args]
        if len(union_args) != len(u1_args_set):
            u1 = self._tensor_unitary(u1, u1_args, union_args)

        if len(union_args) != len(u2_args_set):
            u2 = self._tensor_unitary(u2, u2_args, union_args)

        return self._computer.dot(u2, u1), union_args

    def _tensor_unitary(self, unitary, unitary_args, extend_args):
        uargs_idx = [extend_args.index(uarg) for uarg in unitary_args]
        mono_diff = np.diff(uargs_idx)
        if np.allclose(abs(mono_diff), 1):
            if np.allclose(mono_diff, -1):
                unitary = self._computer.MatrixPermutation(
                    unitary,
                    self._array_helper.arange(len(unitary_args) - 1, -1, -1)
                )

            return self._computer.MatrixTensorI(
                unitary,
                1 << (min(uargs_idx) - 0),
                1 << (len(extend_args) - 1 - max(uargs_idx))
            )

        tensor_matrix = self._computer.MatrixTensorI(
            unitary,
            1,
            1 << (len(extend_args) - len(unitary_args))
        )
        tmatrix_args = unitary_args + [earg for earg in extend_args if earg not in unitary_args]
        permutation_index = [extend_args.index(tm_arg) for tm_arg in tmatrix_args]

        return self._computer.MatrixPermutation(tensor_matrix, self._array_helper.array(permutation_index))

    def _combined_gates(self, gates):
        args_num = [len(args) for _, args in gates]
        while len(args_num) > 1:
            left, right = self._find_mini_args_num(args_num)
            lg, rg = gates[left], gates[right]
            combined_matrix, combined_matrix_args = self.merge_gates(
                lg[0], lg[1],
                rg[0], rg[1]
            )
            gates[left] = (combined_matrix, combined_matrix_args)
            del gates[right]

            args_num[left] += args_num[right]
            del args_num[right]

        return gates[0]

    def _find_mini_args_num(self, nums: list):
        mini_num = sum(nums) * 2
        left, right = -1, -1
        for i in range(1, len(nums), 1):
            if nums[i] + nums[i - 1] < mini_num:
                left = i - 1
                right = i
                mini_num = nums[i] + nums[i - 1]

        return left, right



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/6/9 下午5:35
# @Author  : Kaiqi Li
# @File    : gate_based

import numpy as np
try:
    import cupy as cp
except ImportError:
    cupy = None

from QuICT.core.operator import Operator
from QuICT.core.utils import GateType


class GateMatrixs:
    """
    The class of storing the gates' compute matrix, without duplicate.

    Args:
        precision(Union[np.complex64, np.complex128]): The precision of the gates.
        gpu_device_id(int): The GPU device ID.
    """
    def __init__(self, precision, gpu_device_id: int = 0):
        self.gate_matrixs = {}
        self.precision = precision
        self.device_id = gpu_device_id
        self.matrix_idx = []
        self.matrix_len = 0
        self._unitary_idx = 0

    @property
    def matrix(self):
        """ Return the matrix with all gates' compute matrix. """
        return self.final_matrix

    def _get_gate_name(self, gate):
        if gate.type == GateType.unitary:
            gate_name = gate.name
        else:
            gate_name = str(gate.type)

        for parg in gate.pargs:
            gate_name += f"_{parg}"

        return gate_name

    def build(self, gates):
        for gate in gates:
            if isinstance(gate, Operator):
                continue

            if gate.type in [GateType.measure, GateType.reset, GateType.barrier]:
                continue

            if gate.type == GateType.unitary:
                gate.name = gate.name + f"_{self._unitary_idx}"
                self._unitary_idx += 1

            gate_name = self._get_gate_name(gate)
            if gate_name not in self.gate_matrixs.keys():
                matrix = gate.matrix
                self._build_matrix_gate(gate_name, matrix)

        self._concentrate_gate_matrixs()

    def _build_matrix_gate(self, gate_name, matrix):
        """ Add gate. """
        self.gate_matrixs[gate_name] = (self.matrix_len, matrix.size)
        self.matrix_len += matrix.size
        if matrix.dtype != self.precision:
            matrix = matrix.astype(self.precision)
        self.matrix_idx.append(matrix)

    def _concentrate_gate_matrixs(self):
        """ Combined all gates' computer matrix in one large matrix."""
        self.final_matrix = np.empty(self.matrix_len, dtype=self.precision)
        start = 0

        for matrix in self.matrix_idx:
            self.final_matrix[start:start + matrix.size] = matrix.ravel()[:]
            start += matrix.size

        with cp.cuda.Device(self.device_id):
            self.final_matrix = cp.array(self.final_matrix)

    def get_target_matrix(self, gate):
        """
        Find the compute matrix of the given gate.

        Args:
            gate(Gate): the gate in circuit.
        """
        if gate.type in [GateType.measure, GateType.reset, GateType.barrier]:
            raise KeyError(f"Wrong gate here. {gate.name}")

        gate_name = self._get_gate_name(gate)
        start, itvl = self.gate_matrixs[gate_name]

        return self.final_matrix[start:start + itvl]



import os
import sqlite3


class CircuitLibDB:
    def __init__(self):
        self._file_path = os.path.join(
            os.path.dirname(__file__),
            "../../lib/circuitlib/"
        )
        self._connect = sqlite3.connect(f"{self._file_path}/circuit_library.db")
        self._connect.isolation_level = "EXCLUSIVE"
        self._cursor = self._connect.cursor()

        # Initial Database when first run
        try:
            self._create_table()
        except:
            pass

    def __del__(self):
        self._cursor.close()
        self._connect.close()

    def clean(self):
        """ Clean the table data in current DB. """
        self._cursor.execute("DROP TABLE CIRCUIT_LIB")

    def size(self) -> int:
        self._cursor.execute("SELECT max(rowid) from CIRCUIT_LIB")
        rowcount = self._cursor.fetchone()

        return rowcount[0]

    def circuit_filter(
        self,
        type: str,
        classify: str,
        max_width=None, max_size=None, max_depth=None
    ) -> list:
        """ Get list of qasm file's name which satisfied the condition. """
        based_sql_cmd = "SELECT NAME FROM CIRCUIT_LIB WHERE "
        condition_cmd = f"TYPE=\'{type}\' AND CLASSIFY=\'{classify}\'"
        if isinstance(max_width, int):
            condition_cmd += f" AND WIDTH<=\'{max_width}\'"
        elif isinstance(max_width, list):
            width_str = ", ".join([str(w) for w in max_width])
            condition_cmd += " AND WIDTH IN (%s)" % width_str

        if max_size is not None:
            condition_cmd += f" AND SIZE<=\'{max_size}\'"

        if max_depth is not None:
            condition_cmd += f" AND DEPTH<=\'{max_depth}\'"

        sql_cmd = based_sql_cmd + condition_cmd
        self._cursor.execute(sql_cmd)

        return self._cursor.fetchall()

    def _create_table(self):
        # Temp add here, delete after create datebase
        self._cursor.execute(
            'CREATE TABLE CIRCUIT_LIB(' +
            'ID INTEGER PRIMARY KEY AUTOINCREMENT, NAME text, TYPE text, CLASSIFY text, ' +
            'WIDTH INT, SIZE INT, DEPTH INT)'
        )

    def circuit_exist(self, file_name: str):
        self._cursor.execute(f"SELECT * FROM CIRCUIT_LIB WHERE NAME=\'{file_name}\'")
        file_info = self._cursor.fetchone()

        return file_info is not None

    def add_template_circuit(self):
        file_path = os.path.join(
            self._file_path,
            "template"
        )

        for file in filter(lambda x: x.endswith('.qasm'), os.listdir(file_path)):
            _, width, size, depth, _ = file.split("_")
            width = int(width[1:])
            size = int(size[1:])
            depth = int(depth[1:])

            self._cursor.execute(
                "INSERT INTO CIRCUIT_LIB(NAME, TYPE, CLASSIFY, WIDTH, SIZE, DEPTH)" +
                f"VALUES (\'{file}\', \'template\', \'template\', " +
                f"\'{width}\', \'{size}\', \'{depth}\')"
            )

        self._connect.commit()

    def add_circuit(self, type_: str):
        file_path = os.path.join(
            self._file_path,
            type_
        )

        for classify in os.listdir(file_path):
            if classify == ".keep":
                continue

            folder = os.path.join(file_path, classify)
            for file in filter(lambda x: x.endswith('.qasm'), os.listdir(folder)):
                width, size, depth = file.split("_")[:3]
                width = int(width[1:])
                size = int(size[1:])
                if ".qasm" not in depth:
                    depth = int(depth[1:])
                else:
                    depth = int(depth[1:-5])

                self._cursor.execute(
                    "INSERT INTO CIRCUIT_LIB(NAME, TYPE, CLASSIFY, WIDTH, SIZE, DEPTH)" +
                    f"VALUES (\'{file}\', \'{type_}\', \'{classify}\', " +
                    f"\'{width}\', \'{size}\', \'{depth}\')"
                )

        self._connect.commit()



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/1/1 8:11 下午
# @Author  : Han Yu
# @File    : cnot_local_force.py

from itertools import combinations

from .cnot_force import CnotForceBfs
from .cnot_store_force import CnotStoreForceBfs
from QuICT.core import Circuit
from QuICT.core.gate import CX, GateType
from QuICT.qcda.utility import OutputAligner


class CnotLocalForceBfs(object):
    """
    use bfs to optimize the cnot circuit
    """
    def __init__(self, store=False):
        """
        Args:
            store(bool): whether work with local data
        """
        self.store = store

    @OutputAligner()
    def execute(self, circuit: Circuit):
        """ optimize the circuit locally

        Args:
            circuit(Circuit): the circuit to be optimize

        Returns:
            Circuit: optimal circuit
        """
        gates = circuit.gates
        width = circuit.width()
        last_length = len(gates)
        while True:
            gates = self.traverse(gates, width, self.store)
            new_length = len(gates)
            if last_length == new_length:
                break
            last_length = new_length
        circuit_opt = Circuit(width)
        circuit_opt.extend(gates)
        return circuit_opt

    def traverse(self, input: list, width, store):
        """ find the best circuit by bfs

        Args:
            input(list<BasicGate>): input circuit
            width(int): circuit_width
            store(bool): whether work with local data

        Returns:
            list<BasicGate>: gates after optimization
        """
        all_list = [i for i in range(width)]
        max_local_qubits = 5 if store else 4
        for comb in combinations(all_list, min(width, max_local_qubits)):
            input = self.traverse_with_fix_qubits(input, set(comb), store)
        return input

    def traverse_with_fix_qubits(self, gates: list, fix: set, store):
        """ local optimize for fix qubits

        Args:
            gates(list<CXGate>): the gates to be optimized
            fix(set<int>): the fix qubits
            store(bool): whether work with local data

        Returns:
            list<CXGate>: results after optimization
        """
        mapping = {}
        back_map = {}
        mapping_id = 0
        for element in fix:
            mapping[element] = mapping_id
            back_map[mapping_id] = element
            mapping_id += 1
        output = []
        local_list = []
        for gate in gates:
            if gate.type == GateType.cx:
                fix_in = int(gate.carg in fix) + int(gate.targ in fix)
                if fix_in == 2:
                    local_list.append(gate)
                    continue
                elif fix_in == 0:
                    output.append(gate)
                    continue
            else:
                commu = True
                for s in local_list:
                    if not s.commutative(gate):
                        commu = False
                        break
                if commu:
                    output.append(gate)
                    continue
            if len(local_list) != 0:
                circuit = Circuit(len(fix))
                for local_gate in local_list:
                    CX | circuit([mapping[local_gate.carg], mapping[local_gate.targ]])
                new_circuit = (CnotStoreForceBfs() if store else CnotForceBfs()).execute(circuit)
                for local_gate in new_circuit.gates:
                    new_gate = CX.copy()
                    new_gate.cargs = [back_map[local_gate.carg]]
                    new_gate.targs = [back_map[local_gate.targ]]
                    output.append(new_gate)
                local_list = []
            output.append(gate)

        if len(local_list) != 0:
            circuit = Circuit(len(fix))
            for local_gate in local_list:
                CX | circuit([mapping[local_gate.carg], mapping[local_gate.targ]])
            new_circuit = (CnotStoreForceBfs() if store else CnotForceBfs()).execute(circuit)
            for local_gate in new_circuit.gates:
                new_gate = CX.copy()
                new_gate.cargs = [back_map[local_gate.carg]]
                new_gate.targs = [back_map[local_gate.targ]]
                output.append(new_gate)

        return output



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/2/11 10:30
# @Author  : Han Yu
# @File    : __init__.py.py

from ._algorithm import Algorithm
from .quantum_algorithm import *
from .quantum_machine_learning import *
from .weight_decision import WeightDecision



from __future__ import annotations

import os.path as osp
from collections import deque
from random import randint, sample
from typing import List, Optional, Tuple, Union

import networkx as nx
import torch
from torch_geometric.data import Batch as PygBatch
from torch_geometric.data import Data as PygData

from QuICT.core import *
from QuICT.core.gate import CompositeGate, GateType
from QuICT.qcda.mapping.common.circuit_info import CircuitInfo as CircuitInfoBase
from QuICT.qcda.mapping.common.data_factory import DataFactory as DataFactoryBase
from QuICT.qcda.mapping.common.layout_info import LayoutInfo as LayoutInfoBase


class DataFactory(DataFactoryBase):
    """DataFactory is used to create random data for training."""

    def __init__(
        self, topo: Union[str, Layout], max_gate_num: int, data_dir: str = None
    ) -> None:
        if data_dir is None:
            data_dir = osp.abspath(osp.dirname(__file__))
            data_dir = osp.join(data_dir, "data")
        super().__init__(topo, max_gate_num, data_dir)

    def get_one(self) -> "State":
        topo = self._cur_topo
        qubit_num = self.topo_qubit_num_map[topo.name]
        topo_graph = self.topo_graph_map[topo.name]

        min_gn = self._max_gate_num // 3
        gate_num = randint(min_gn, max(self._max_gate_num, min_gn))
        success = False
        while not success:
            circ = Circuit(qubit_num)
            circ.random_append(
                gate_num,
                typelist=[
                    GateType.crz,
                ],
            )
            circ_state = CircuitInfo(circ=circ, max_gate_num=self._max_gate_num)
            logic2phy = [i for i in range(self.topo_qubit_num_map[topo.name])]
            circ_state.eager_exec(logic2phy=logic2phy, topo_graph=topo_graph)
            success = circ_state.count_gate() > 0

        state = State(
            circ_info=circ_state,
            layout=topo,
            logic2phy=logic2phy,
        )
        return state


class CircuitInfo(CircuitInfoBase):
    """DAG Representation of a quantum circuit. A virtual node will be
    added with label 0.
    """

    def layered_gate_matrices(self, logic2phy: List[int]) -> torch.Tensor:
        layers: List[List[Tuple[int, int]]] = []
        occupied = [-1 for _ in range(self._qubit_num)]
        for gid in sorted(self._graph.nodes):
            gate = self._gates[gid]
            a, b = gate.cargs + gate.targs
            layer_id = 1 + max(occupied[a], occupied[b])
            while len(layers) <= layer_id:
                layers.append([])
            occupied[a] = layer_id
            occupied[b] = layer_id
            _a, _b = logic2phy[a], logic2phy[b]
            layers[layer_id].append((_a, _b))
        q = self._qubit_num
        ans = torch.zeros(len(layers), q, q)
        for idx, layer in enumerate(layers):
            for u, v in layer:
                ans[idx][u][v] = 1
                ans[idx][v][u] = 1
        ans = ans.view(-1, q * q)
        return ans

    def remained_circ(self, logic2phy: List[int]) -> CompositeGate:
        gates = {}
        for bit_stick in self._bit2gid:
            for gid in bit_stick:
                if gid not in gates:
                    gates[gid] = self._gates[gid]
        cg = CompositeGate()
        for gate in gates.values():
            a, b = gate.cargs + gate.targs
            a, b = logic2phy[a], logic2phy[b]
            with cg:
                gate & [a, b]
        return cg

    def to_pyg(self, logic2phy: List[int]) -> PygData:
        """Convert current data into PyG Data according to current mapping.

        Arg:
            logic2phy (List[int]): Logical to physical qubit mapping.
            max_qubit_num (int): Padding size.

        Returns:
            PygData: PyG data.
        """
        x = torch.zeros(self._max_gate_num, 2, dtype=torch.long)
        edge_index = []
        g: nx.DiGraph = nx.convert_node_labels_to_integers(self._graph)
        assert g.number_of_nodes() > 0
        # g = self._graph
        for node in g.nodes:
            gid = g.nodes[node]["gid"]
            gate = self._gates[gid]
            args = gate.cargs + gate.targs
            if len(args) == 1:
                continue
            a, b = args
            x[node][0] = logic2phy[a] + 1
            x[node][1] = logic2phy[b] + 1
            # Add self loops
            edge_index.append([gid, gid])
        for u, v in g.edges:
            edge_index.append([u, v])
        edge_index = torch.tensor(edge_index, dtype=torch.long).T.contiguous()
        data = PygData(x=x, edge_index=edge_index)
        return data


class LayoutInfo(LayoutInfoBase):
    def __init__(self, layout: Layout) -> None:
        super().__init__(layout)


class StateSlim:
    """Intermediate state for an actor, remained only core information."""

    def __init__(
        self,
        circ_info: PygData,
        logic2phy: List[int],
        phy2logic: Optional[List[int]] = None,
    ) -> None:
        self.circ_info = circ_info
        self.logic2phy = logic2phy
        """Logical to physical mapping
        """
        q = len(logic2phy)
        self.phy2logic = phy2logic
        if self.phy2logic is None:
            self.phy2logic = [0 for _ in range(q)]
            for i in range(q):
                self.phy2logic[self.logic2phy[i]] = i

        self.logic2phy = torch.tensor(self.logic2phy)
        self.phy2logic = torch.tensor(self.phy2logic)

    def to_nn_data(self):
        return self.circ_info

    @staticmethod
    def batch_from_list(data_list: list, device: str):
        return PygBatch.from_data_list(data_list=data_list).to(device=device)

    def to_tensor_list(self) -> List[torch.Tensor]:
        return [
            self.circ_info.x,
            self.circ_info.edge_index,
            self.logic2phy,
            self.phy2logic,
        ]

    @staticmethod
    def from_tensor_list(t_list: List[torch.Tensor]) -> StateSlim:
        return StateSlim(
            circ_info=PygData(x=t_list[0], edge_index=t_list[1]),
            logic2phy=t_list[2],
            phy2logic=t_list[3],
        )


class State:
    """Intermediate state for an actor."""

    def __init__(
        self,
        circ_info: CircuitInfo,
        layout: Layout,
        logic2phy: List[int],
        phy2logic: Optional[List[int]] = None,
    ) -> None:
        self.circ_info = circ_info
        self.layout_info = LayoutInfo(layout)
        self.logic2phy = logic2phy
        """Logical to physical mapping
        """
        q = layout.qubit_number
        self.phy2logic = phy2logic
        if self.phy2logic is None:
            self.phy2logic = [0 for _ in range(q)]
            for i in range(q):
                self.phy2logic[self.logic2phy[i]] = i

        self._circ_layered_matrices = None
        self._slim = None

    @property
    def circ_layered_matrices(self) -> torch.Tensor:
        if self._circ_layered_matrices is None:
            self._circ_layered_matrices = self.circ_info.layered_gate_matrices(
                self.logic2phy
            )
        return self._circ_layered_matrices

    def remained_circ(self) -> CompositeGate:
        return self.circ_info.remained_circ(self.logic2phy)

    def biased_random_swap(self, exclude: tuple) -> Tuple[int, int]:
        return self.circ_info.biased_random_swap(
            self.layout_info.topo_dist, self.logic2phy, exclude
        )

    def eager_exec(
        self,
        physical_circ: Optional[CompositeGate] = None,
    ) -> int:
        self._slim = None
        return self.circ_info.eager_exec(
            self.logic2phy,
            self.layout_info.topo_graph,
            physical_circ,
        )

    def to_slim(self) -> StateSlim:
        if self._slim is None:
            self._slim = StateSlim(
                circ_info=self.circ_info.to_pyg(self.logic2phy),
                logic2phy=self.logic2phy,
                phy2logic=self.phy2logic,
            )
        return self._slim


class Transition:
    """States before and after certain action, together with corresponding reward."""

    def __init__(
        self,
        state: StateSlim,
        action: Tuple[int, int],
        next_state: StateSlim,
        reward: float,
    ) -> None:
        self.state = state
        self.action = action
        self.next_state = next_state
        self.reward = reward

    def __iter__(self):
        return iter((self.state, self.action, self.next_state, self.reward))


class ReplayMemory:
    """Bounded buffer of transitions."""

    def __init__(self, capacity: int) -> None:
        self._memory = deque([], maxlen=capacity)

    def push(self, transition: Transition):
        self._memory.append(transition)

    def sample(self, batch_size: int) -> List[Transition]:
        return sample(self._memory, batch_size)

    def __len__(self) -> int:
        return len(self._memory)

    def __iter__(self):
        return iter(self._memory)

    def clear(self):
        self._memory.clear()


class ValidationData:
    """Container of some circuit generated by other algorithms. Used for validation during training."""

    def __init__(
        self,
        circ: Circuit,
        topo: Layout,
        rl_mapped_circ: Union[Circuit, CompositeGate] = None,
        remained_circ: Union[Circuit, CompositeGate] = None,
    ) -> None:
        self.circ = circ
        self.rl_mapped_circ = rl_mapped_circ
        self.topo = topo
        self.remained_circ = remained_circ


class TrainConfig:
    def __init__(
        self,
        topo: Union[str, Layout],
        max_gate_num: int = 300,
        feat_dim: int = 150,
        gamma: float = 0.95,
        replay_pool_size: int = 10_000_000,
        lr: float = 0.0001,
        batch_size: int = 64,
        total_epoch: int = 2000,
        explore_period: int = 10000,
        target_update_period: int = 20,
        actor_num: int = 2,
        world_size: int = 3,
        model_sync_period: int = 10,
        memory_sync_period: int = 10,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
        model_path: str = None,
        log_dir: str = None,
        epsilon_start: float = 0.95,
        epsilon_end: float = 0.05,
        epsilon_decay: float = 5_000_000.0,
        reward_scale: float = 15.0,
        inference: bool = False,
        inference_model_dir: str = "./model",
    ) -> None:
        self.factory = DataFactory(topo=topo, max_gate_num=max_gate_num)

        self.topo = self.factory._cur_topo

        swaps = [(edge.u, edge.v) for edge in self.topo]
        swaps.sort()
        self.action_id_by_swap = {}
        self.swap_by_action_id = {}
        for idx, swap in enumerate(swaps):
            self.action_id_by_swap[swap] = idx
            self.swap_by_action_id[idx] = swap
        self.action_num = len(self.action_id_by_swap)

        self.distributed_backend = "nccl" if "cuda" in device else "gloo"

        self.max_gate_num = max_gate_num
        self.feat_dim = feat_dim
        self.gamma = gamma
        self.replay_pool_size = replay_pool_size
        self.lr = lr
        self.batch_size = batch_size
        self.total_epoch = total_epoch
        self.explore_period = explore_period
        self.target_update_period = target_update_period
        self.model_sync_period = model_sync_period
        self.memory_sync_period = memory_sync_period
        self.actor_num = actor_num
        self.world_size = world_size
        self.device = device
        self.epsilon_start = epsilon_start
        self.epsilon_end = epsilon_end
        self.epsilon_decay = epsilon_decay
        self.reward_scale = reward_scale
        self.inference = inference
        self.inference_model_dir = inference_model_dir

        if model_path is None:
            model_path = osp.dirname(osp.abspath(__file__))
            model_path = osp.join(model_path, f"{self.topo.name}-model_rl_mapping")
        self.model_path = model_path

        if log_dir is None:
            log_dir = osp.dirname(osp.abspath(__file__))
            log_dir = osp.join(log_dir, "torch_runs")
            log_dir = osp.join(log_dir, "rl_mapping")
        self.log_dir = log_dir



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/3/30 6:24
# @Author  : Han Yu
# @File    : __init__.py.py

from .local_drawer import PhotoDrawer
from .text_drawer import TextDrawing
from .tikz_drawer import TikzDrawer



import numpy as np

from QuICT.core import Circuit
from QuICT.core.gate import *
from QuICT.core.gate.backend import MCTOneAux
from QuICT.tools import Logger
from QuICT.tools.exception.core import *

logger = Logger("Grover")

ALPHA = 1.5


def degree_counterclockwise(v1: np.ndarray, v2: np.ndarray):
    """from v1 to v2
    """
    d = np.real(np.arccos(sum(v1 * v2) / np.sqrt(sum(v1 * v1) * sum(v2 * v2))))
    if d > 0.5 * np.pi:
        d = np.pi - d
    return d


class Grover:
    """ simple grover

    Quantum Computation and Quantum Information - Michael A. Nielsen & Isaac L. Chuang
    """

    def __init__(self, simulator) -> None:
        self.simulator = simulator

    def _grover_operator(self, n, n_ancilla, oracle, is_bit_flip=False):
        cgate = CompositeGate()
        index_q = list(range(n))
        ancilla_q = list(range(n, n + n_ancilla))
        # Grover iteration
        if is_bit_flip:
            X | cgate(ancilla_q[0])
            H | cgate(ancilla_q[0])
        oracle | cgate(index_q + ancilla_q)
        if is_bit_flip:
            H | cgate(ancilla_q[0])
            X | cgate(ancilla_q[0])
        for idx in index_q:
            H | cgate(idx)
        # control phase shift
        for idx in index_q:
            X | cgate(idx)
        H | cgate(index_q[n - 1])
        MCTOneAux().execute(n + 1) | cgate(index_q + ancilla_q[:1])
        H | cgate(index_q[n - 1])
        for idx in index_q:
            X | cgate(idx)
        # control phase shift end
        for idx in index_q:
            H | cgate(idx)
        return cgate

    def circuit(
        self,
        n,
        n_ancilla,
        oracle,
        n_solution=1,
        measure=True,
        is_bit_flip=False,
        iteration_number_forced=False,
    ):
        """ grover search for f with custom oracle

        Args:
            n(int): the length of input of f
            n_ancilla(int): length of oracle working space. assume clean
            oracle(CompositeGate): the oracle that flip phase of target state.
                [0:n] is index qreg,
                [n:n+k] is ancilla
            n_solution(int): number of solution
            measure(bool): measure included or not
            iteration_number_forced(bool): if True, n_solution is used as iteration count

        Returns:
            int: the a satisfies that f(a) = 1
        """
        assert n_ancilla > 0, "at least 1 ancilla, which is shared by MCT part"
        circuit = Circuit(n + n_ancilla)
        index_q = list(range(n))
        if iteration_number_forced:
            T = n_solution
        else:
            N = 2 ** n
            theta = np.arcsin(np.sqrt(n_solution / N))
            T = int(np.round((np.pi / 2 - theta) / (2 * theta)))

        grover_operator = self._grover_operator(n, n_ancilla, oracle, is_bit_flip)

        # create equal superposition state in index_q
        for idx in index_q:
            H | circuit(idx)
        # rotation
        for i in range(T):
            grover_operator | circuit
        for idx in index_q:
            if measure:
                Measure | circuit(idx)
        logger.info(
            f"circuit width          = {circuit.width():4}\n"
            + f"oracle  calls          = {T:4}\n"
            + f"other circuit size     = {circuit.size() - oracle.size()*T:4}\n"
        )
        return circuit

    def run(
        self,
        n,
        n_ancilla,
        oracle,
        n_solution=1,
        measure=True,
        is_bit_flip=False,
        check_solution=None,
    ):
        simulator = self.simulator
        index_q = list(range(n))
        # unkonwn solution number
        if n_solution is None:
            assert check_solution is not None
            n_solution_guess = 1 << n
            while n_solution_guess > 0:
                logger.info(f"trial with {n_solution_guess} solutions...")
                circ = self.circuit(
                    n, n_ancilla, oracle, n_solution_guess, True, is_bit_flip
                )
                simulator.run(circ)
                solution = int(circ[index_q])
                if check_solution(solution):
                    return solution
                n_solution_guess = int(n_solution_guess / ALPHA)
            logger.info("FAILED!")
            return None
        # no solution
        elif n_solution == 0:
            return 0
        # standard Grover's algorithm
        else:
            circ = self.circuit(n, n_ancilla, oracle, n_solution, measure, is_bit_flip)
            simulator.run(circ)
            return int(circ[index_q])



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/9/11 08:06
# @Author  : Xiaoquan Xu
# @File    : moleculardata.py

import h5py


def get_from_file(item, molfile, set_type=None):
    """
    Args:
        item(str): the name of the target dataset
        molfile(dict): the read pointer of the molecular file
        set_type(str): the target type of data
    """
    try:
        data = molfile[item][...]
        if data.dtype.num == 0:
            data = None
        elif set_type is not None:
            data = data.astype(set_type)
    except Exception:
        data = None
        print("!! no " + item)
    return data


class MolecularData:
    """
    Attributes:
        molfile(str): the molecular file name with ".hdf5" suffix
        n_orbitals(integer): the number of the spatial orbitals
        n_electrons(integer): the number of the electrons in the molecule
        nuclear_repulsion(float): the energy from nuclei-nuclei interaction.
        one_body_integrals(ndarray): the one-electron integrals
            in the shape of (n_orbitals, n_orbitals)
        two_body_integrals(ndarray): the two-electron integrals
            in the shape of (n_orbitals, n_orbitals, n_orbitals, n_orbitals)
    """
    def __init__(self, molfile=""):
        """
        Args:
            molfile(str): the molecular file name
        """
        molfile = molfile.strip()
        if molfile[-5:] != ".hdf5":
            self.molfile = molfile + ".hdf5"
        else:
            self.molfile = molfile

        with h5py.File(self.molfile, "r") as f:
            self.n_orbitals = get_from_file("n_orbitals", f, 'int')
            self.n_electrons = get_from_file("n_electrons", f, 'int')
            self.nuclear_repulsion = get_from_file("nuclear_repulsion", f, 'float')
            self.one_body_integrals = get_from_file("one_body_integrals", f)
            self.two_body_integrals = get_from_file("two_body_integrals", f)

    def get_integrals(self):
        return self.one_body_integrals, self.two_body_integrals

    def save(self):
        with h5py.File("save_" + self.molfile, "w") as f:
            f.create_dataset("n_orbitals", data=str(self.n_orbitals))
            f.create_dataset("n_electrons", data=str(self.n_electrons))
            f.create_dataset("nuclear_repulsion", data=str(self.nuclear_repulsion))
            f.create_dataset("one_body_integrals", data=str(self.one_body_integrals))
            f.create_dataset("two_body_integrals", data=str(self.two_body_integrals))



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/8/17 17:30
# @Author  : Xiaoquan Xu
# @File    : encoder.py

"""
Encoders transform ladder operators to CompositeGate on zero state of qubits.
"""

from QuICT.chemistry.operator.qubit_operator import QubitOperator


class Encoder:
    """
    Superclass of encoding methods.

    Attributes:
        n_orbitals(integer): the number of orbitals
    """
    def __init__(self, n_orbitals):
        self.n_orbitals = n_orbitals

    def encoder(self, fermion_operator):
        """
        Encoders transform ladder operators to Qubit Operators.

        Args:
            fermion_operator(FermionOperator): FermionOperator to be transformed

        Returns:
            QubitOperator: The corresponding operators on qubits
        """
        n_orbitals = self.n_orbitals
        ans = QubitOperator()
        for mono_f in fermion_operator.operators:
            mono_q = QubitOperator([], mono_f[1])
            for operator in mono_f[0]:
                mono_q *= self.encoder_single(operator[0], operator[1], n_orbitals)
            ans += mono_q
        return ans

    def encoder_single(self, target, kind, n_orbitals):
        """
        Encode a single ladder operator (To be overrided).

        Args:
            target(integer): the target fermion of the single ladder operator (>=0)
            kind(integer): whether the operator is annihilation(1) or creation(0)
            n_orbitals(integer): the number of all the orbitals

        Returns:
            QubitOperator: The corresponding operators on qubits
        """
        raise Exception("The encoder is not realized.")


def trans_01(target):
    """
    Construct the operator |0><1| on the target qubit.
    The next three functions are similar.

    Args:
        target(integer): index of the target qubit (>=0)

    Returns:
        QubitOperator: '0.5X+0.5iY' on the target qubit
    """
    return QubitOperator([(target, 1)], 0.5) + QubitOperator([(target, 2)], 0.5j)


def trans_10(target):
    return QubitOperator([(target, 1)], 0.5) + QubitOperator([(target, 2)], -0.5j)


def trans_00(target):
    return QubitOperator([], 0.5) + QubitOperator([(target, 3)], 0.5)


def trans_11(target):
    return QubitOperator([], 0.5) + QubitOperator([(target, 3)], -0.5)


class JordanWigner(Encoder):
    """
    Implement the Jordan-Wigner encoding method.
    """
    def __init__(self, n_orbitals=None):
        super().__init__(n_orbitals)

    def encoder_single(self, target, kind, n_orbitals):
        Zlist = [(i, 3) for i in range(target)]
        ans = QubitOperator(Zlist)
        # annihilation
        if kind == 0:
            ans *= trans_01(target)
        # creation
        else:
            ans *= trans_10(target)
        return ans


class Parity(Encoder):
    """
    Implement the parity encoding method.
    """
    def encoder_single(self, target, kind, n_orbitals):
        Xlist = [(i, 1) for i in range(target + 1, n_orbitals)]
        ans = QubitOperator(Xlist)
        # annihilation
        if kind == 0:
            if target == 0:
                ans *= trans_01(target)
            else:
                ans *= trans_00(target - 1) * trans_01(target) - trans_11(target - 1) * trans_10(target)
        # creation
        else:
            if target == 0:
                ans *= trans_10(target)
            else:
                ans *= trans_00(target - 1) * trans_10(target) - trans_11(target - 1) * trans_01(target)
        return ans


def flip(x, n_orbitals):
    """
    Construct a list of indexes involved in flipping n_x.

    Args:
        x(integer): index of the target qubit (x>=0)
        n_orbitals(integer): the number of all the orbitals

    Returns:
        list: the indexes involved in flipping n_x
    """
    index = []
    x += 1
    while (x <= n_orbitals):
        index.append(x - 1)
        x += x & (-x)
    return index


def sumup(x):
    """
    Construct a list of indexes involved in summation n_0, ..., n_x

    Args:
        x(integer): index of the target qubit. (x>=0)
            Specially, if x is -1, return []

    Returns:
        list: the indexes involved in summation n_0, ..., n_x
    """
    index = []
    x += 1
    while (x > 0):
        index.append(x - 1)
        x -= x & (-x)
    return index[::-1]


class BravyiKitaev(Encoder):
    """
    Implement the Bravyi-Kitaev encoding method
    """
    def encoder_single(self, target, kind, n_orbitals):
        Xlist = [(i, 1) for i in flip(target, n_orbitals)]
        ans = QubitOperator(Xlist, 0.5)
        Zlist1 = [(i, 3) for i in sumup(target - 1)]
        Zlist2 = [(i, 3) for i in sumup(target)]
        ans *= QubitOperator(Zlist1) - QubitOperator(Zlist2, (-1)**kind)
        return ans



"""
Some codes below are from pybind11 cmake example:
https://github.com/pybind/cmake_example/blob/0baee7e073a9b3738052f543e6bed412aaa22750/setup.py
"""

import os
import platform
import subprocess
import sys
from os import getcwd, path
from typing import List, Tuple

import pybind11
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

pybind11_cmake_dir = pybind11.__path__[0]
for p in ["share", "cmake", "pybind11"]:
    pybind11_cmake_dir = path.join(pybind11_cmake_dir, p)

# print helpers
def print_segment():
    print("\033[92m", "=" * 80, "\033[39m", sep="")


def print_cyan(segment):
    print(f"\033[36m{segment}\033[39m")


def print_magenta(segment):
    print(f"\033[95m{segment}\033[39m")


def print_yellow(segment):
    print(f"\033[33m{segment}\033[39m")


def print_if_not_none(segment):
    if segment:
        print(segment)


def print_with_wrapper(header, out_obj):
    if header[0] != "\033":
        if len(header) > 12:
            header = header[:9] + "..."
        if len(header) < 12:
            for _ in range(12 - len(header)):
                header += "."

        header = f"\033[36m[{header}]\033[39m "

    if out_obj is None:
        return
    if isinstance(out_obj, str):
        print(header, out_obj)
    else:
        for line in iter(out_obj.readline, b""):
            print(header, line.decode("utf-8"), sep="", end="")


def run_with_output_wrapper(header, args, cwd, shell=(platform.system() == "Windows")):
    if len(header) > 12:
        header = header[:9] + "..."
    if len(header) < 12:
        for _ in range(12 - len(header)):
            header += "."

    header = f"\033[36m[{header}]\033[39m "

    with subprocess.Popen(
        args=args,
        cwd=cwd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        # In Linux, execute in shell cause cmake errors...
        shell=shell,
        # universal_newlines=True,
    ) as proc:
        print_with_wrapper(header, proc.stdout)
        ret_code = proc.wait()
    if ret_code:
        raise subprocess.CalledProcessError(ret_code, args)


# Detect if I'm in `root` or `root/build`
PY_FILE_PATH = path.dirname(path.abspath(__file__))

PRJ_ROOT_RELATIVE = "." if getcwd() == PY_FILE_PATH else ".."
PRJ_ROOT = path.abspath(PRJ_ROOT_RELATIVE)

# Convert distutils Windows platform specifiers to CMake -A arguments
PLAT_TO_CMAKE = {
    "win32": "Win32",
    "win-amd64": "x64",
    "win-arm32": "ARM",
    "win-arm64": "ARM64",
}


# A CMakeExtension needs a source dir instead of a file list.
# The name must be the _single_ output extension from the CMake build.
class CMakeExtension(Extension):
    def __init__(self, name, source_dir, extra_cmake_macro=None):
        Extension.__init__(self, name, sources=[])
        self.source_dir = os.path.abspath(source_dir)
        self.extra_cmake_macro = extra_cmake_macro


class ExtensionBuild(build_ext):
    def build_extension(self, ext):
        if isinstance(ext, CMakeExtension):
            self.cmake_build_extension(ext)

    def prepare_cmake_args(
        self, cmake_generator: str, ext_dir: str, cfg: str
    ) -> Tuple[List[str], List[str]]:
        configure_args = []
        build_args = []
        # Set Python_EXECUTABLE instead if you use PYBIND11_FINDPYTHON
        # EXAMPLE_VERSION_INFO shows you how to pass a value into the C++ code
        # from Python.
        configure_args = [
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={ext_dir}",
            f"-DPYTHON_EXECUTABLE={sys.executable}",
            f'-DBUILD_VERSION_INFO="{self.distribution.get_version()}"',
            f"-DCMAKE_BUILD_TYPE={cfg}",  # not used on MSVC, but no harm
            f"-Dpybind11_DIR={pybind11_cmake_dir}",
        ]
        build_args = []

        if self.compiler.compiler_type != "msvc":
            if not cmake_generator:
                configure_args += ["-GUnix Makefiles"]
        else:
            # Ensure CC/CXX is set. This is a fix for Windows PowerShell
            if "CC" in os.environ:
                configure_args += [
                    f"\"-DCMAKE_C_COMPILER:FILEPATH={os.environ['CC']}\""
                ]
            if "CXX" in os.environ:
                configure_args += [
                    f"\"-DCMAKE_CXX_COMPILER:FILEPATH={os.environ['CXX']}\""
                ]
            # Single config generators are handled "normally"
            single_config = any(x in cmake_generator for x in ("NMake", "Ninja"))

            # CMake allows an arch-in-generator style for backward compatibility
            contains_arch = any(x in cmake_generator for x in ("ARM", "Win64"))

            # Specify the arch if using MSVC generator, but only if it doesn't
            # contain a backward-compatibility arch spec already in the
            # generator name.
            if not single_config and not contains_arch:
                configure_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]

            # Multi-config generators have a different way to specify configs
            if not single_config:
                configure_args += [
                    f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{cfg.upper()}={ext_dir}"
                ]
                build_args += ["--config", cfg]
            # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
            # across all generators.
            if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
                # self.parallel is a Python 3 only way to set parallel jobs by hand
                # using -j in the build_ext call, not supported by pip or PyPA-build.
                if hasattr(self, "parallel") and self.parallel:
                    # CMake 3.12+ only.
                    build_args += [f"-j{self.parallel}"]
        return configure_args, build_args

    def cmake_build_extension(self, ext):
        ext_dir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

        # required for auto-detection of auxiliary "native" libs
        if not ext_dir.endswith(os.path.sep):
            ext_dir += os.path.sep

        cfg = "Debug" if self.debug else "Release"

        # CMake lets you override the generator - we need to check this.
        # Can be set with Conda-Build, for example.
        cmake_generator = os.environ.get("CMAKE_GENERATOR", "")

        configure_args, build_args = self.prepare_cmake_args(
            cmake_generator, ext_dir, cfg
        )

        ext_name = ext.name
        if ext_name[-1] == ".":
            ext_name = ext_name[:-1]
        build_temp = f"{self.build_temp}.{ext_name}"
        ext_name_split = ext_name.split(".")
        ext_name = ext_name_split[-1]

        if not os.path.exists(build_temp):
            os.makedirs(build_temp)

        if hasattr(self, "parallel") and self.parallel:
            print_yellow(
                "Extensions are built in parallel. Shell output might be messed up."
            )
        print_cyan(f"[{ext_name}]")
        cmake_cmd = ["cmake"] + configure_args + [f"-S{ext.source_dir}"]
        print_with_wrapper(ext_name, " ".join(cmake_cmd))
        print_with_wrapper(ext_name, "Configuring...")
        run_with_output_wrapper(
            header=ext_name,
            args=cmake_cmd,
            cwd=build_temp,
        )
        print_with_wrapper(ext_name, "Building...")
        run_with_output_wrapper(
            header=ext_name,
            args=["cmake", "--build", "."] + build_args,
            cwd=build_temp,
        )
        libs = []
        for file in os.listdir(ext_dir):
            if file.endswith(".so") or file.endswith(".pyd"):
                libs.append(f"{ext_dir}{file}")
        print_with_wrapper(ext_name, f"Copying back {libs}...")
        run_with_output_wrapper(
            header=ext_name,
            args=["cp", " ".join(libs), ext.source_dir],
            cwd=build_temp,
        )


# static file
file_data = [
    ("QuICT/lib/qasm/libs", [f"{PRJ_ROOT_RELATIVE}/QuICT/lib/qasm/libs/qelib1.inc"]),
    ("QuICT/simulation/utils", [f"{PRJ_ROOT_RELATIVE}/QuICT/simulation/utils/simulator_parameters.json"])
]

setup(
    name="quict",
    version="0.9.1",
    description="Quantum Compute Platform of Institute of Computing Technology",
    author="Library for Quantum Computation and Theoretical Computer Science, ICT, CAS",
    author_email="likaiqi@ict.ac.cn",
    license="Apache License 2.0",
    platforms=["Windows", "Linux", "macOS"],
    url="https://e.gitee.com/quictucas/repos/quictucas/quict",
    package_dir={"QuICT": f"{PRJ_ROOT_RELATIVE}/QuICT"},
    entry_points={
        "console_scripts": [
            "quict = QuICT.tools.cli.quict:main",
        ],
    },
    install_requires=[
        "contourpy==1.0.5",
        "cycler==0.11.0",
        "fonttools==4.37.4",
        "kiwisolver==1.4.4",
        "llvmlite==0.39.1",
        "matplotlib==3.6.1",
        "networkx==2.8.7",
        "numba==0.56.3",
        "numpy==1.23.4",
        "packaging==21.3",
        "Pillow==9.2.0",
        "ply==3.11",
        "pybind11==2.10.0",
        "pyparsing==3.0.9",
        "python-dateutil==2.8.2",
        "scipy==1.9.2",
        "six==1.16.0",
        "ujson==5.5.0",
    ],
    ext_modules=[
        CMakeExtension(
            "QuICT.simulation.state_vector.cpu_simulator.",
            f"{PRJ_ROOT}/QuICT/simulation/state_vector/cpu_simulator/",
        ),
    ],
    cmdclass={"build_ext": ExtensionBuild},
    packages=find_packages(where=PRJ_ROOT_RELATIVE),
    data_files=file_data,
    include_package_data=True,
    python_requires=">=3.8",
    zip_safe=False,
)
print_segment()



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/4/3 13:12
# @Author  : Li Haomin
# @File    : hrs.py

from QuICT.core import Circuit
from QuICT.core.gate import *
from QuICT.algorithm.quantum_algorithm.shor.utility import int2bitwise, mod_reverse


def var_controlled_X(gate_set, controls, target):
    """0 to 2 controlled-X

    Args:
        controls (list): indices of control qubits
        target (int): target
    """
    n_control = len(controls)
    with gate_set:
        if n_control == 0:
            X & target
        elif n_control == 1:
            CX & [controls[0], target]
        elif n_control == 2:
            CCX.build_gate() & [controls[0], controls[1], target]
        else:
            raise ValueError()


def carry(gate_set, control, a, c_bitwise, g_aug, overflow):
    """
    Compute the overflow of a(quantum)+c(classical) with borrowed qubits g_aug.
    at most 2 control bits can be used

    Args:
        control(list): indices of 0 to 2 qubits.
        a(list): n qubits.
        g_aug(list): indices of n-1 qubits(more bits are OK).
        overflow(int): index of 1 qubit.
        c_bitwise(char array): n bits '0'-'1' array, representing binary int c.
    """
    n = len(a)
    n_control = len(control)
    g = g_aug[0:n - 1]
    # n==1, no borrowed bits g
    with gate_set:
        if n == 1:
            if c_bitwise[0] == "1":
                if n_control == 0:
                    CX & [a[0], overflow]
                elif n_control == 1:
                    CCX.build_gate() & [control[0], a[0], overflow]
                elif n_control == 2:
                    # CCCX | (c[0],c[1],a[0],overflow) with g_aug[0] as ancilla
                    CCX.build_gate() & [a[0], g_aug[0], overflow]
                    CCX.build_gate() & [control[0], control[1], g_aug[0]]
                    CCX.build_gate() & [a[0], g_aug[0], overflow]
                    CCX.build_gate() & [control[0], control[1], g_aug[0]]
                else:
                    raise ValueError()
            return
        # n>=2
        if n_control == 0:
            CX & [g[0], overflow]
        elif n_control == 1:
            CCX.build_gate() & [control[0], g[0], overflow]
        elif n_control == 2:
            # CCCX | (c1,c2,g[0],overflow) with a[0] as ancilla
            CCX.build_gate() & [g[0], a[0], overflow]
            CCX.build_gate() & [control[0], control[1], a[0]]
            CCX.build_gate() & [g[0], a[0], overflow]
            CCX.build_gate() & [control[0], control[1], a[0]]
        else:
            raise ValueError()

        for i in range(n - 2):
            if c_bitwise[i] == "1":
                CX & [a[i], g[i]]
                X & a[i]
            CCX.build_gate() & [g[i + 1], a[i], g[i]]
        if c_bitwise[n - 2] == "1":
            CX & [a[n - 2], g[n - 2]]
            X & a[n - 2]
        if c_bitwise[n - 1] == "1":
            CCX.build_gate() & [a[n - 1], a[n - 2], g[n - 2]]
        for i in range(n - 2):
            CCX.build_gate() & [g[n - 2 - i], a[n - 3 - i], g[n - 3 - i]]

        if n_control == 0:
            CX & [g[0], overflow]
        elif n_control == 1:
            CCX.build_gate() & [control[0], g[0], overflow]
        elif n_control == 2:
            # CCCX | (c1,c2,g[0],overflow) with a[0] as ancilla
            CCX.build_gate() & [g[0], a[0], overflow]
            CCX.build_gate() & [control[0], control[1], a[0]]
            CCX.build_gate() & [g[0], a[0], overflow]
            CCX.build_gate() & [control[0], control[1], a[0]]
        else:
            raise ValueError()

        # uncomputation
        for i in range(n - 2):
            CCX.build_gate() & [g[i + 1], a[i], g[i]]
        if c_bitwise[n - 1] == "1":
            CCX.build_gate() & [a[n - 1], a[n - 2], g[n - 2]]
        if c_bitwise[n - 2] == "1":
            X & a[n - 2]
            CX & [a[n - 2], g[n - 2]]
        for i in range(n - 2):
            CCX.build_gate() & [g[n - 2 - i], a[n - 3 - i], g[n - 3 - i]]
            if c_bitwise[n - 3 - i] == "1":
                X & a[n - 3 - i]
                CX & [a[n - 3 - i], g[n - 3 - i]]


def sub_widget(gate_set, v, g):
    """
    sub_widget used in incrementer().

    Args:
        v(list): indices of n qubits.
        g(list): indices of n qubits(more qubits are OK).
    """
    n = len(v)
    if len(g) < n:
        print("When do Sub_Widget, no edequate ancilla qubit")
    with gate_set:
        for i in range(n - 1):
            CX & [g[n - 1 - i], v[n - 1 - i]]
            CX & [g[n - 2 - i], g[n - 1 - i]]
            CCX.build_gate() & [g[n - 1 - i], v[n - 1 - i], g[n - 2 - i]]
        CX & [g[0], v[0]]
        for i in range(n - 1):
            CCX.build_gate() & [g[i + 1], v[i + 1], g[i]]
            CX & [g[i], g[i + 1]]
            CX & [g[i], v[i + 1]]


def incrementer(gate_set, v, g):
    """
    Incremente v by 1, with borrowed qubits g.

    Args:
        v(list): indices of n qubits.
        g(list): indices of n qubits(more qubits are OK).
    """
    n = len(v)
    if len(g) < n:
        print("When do Increment, no edequate borrowed qubit")
    with gate_set:
        for i in range(n):
            CX & [g[n - 1], v[i]]
        for i in range(n - 1):
            X & g[i]
        X & v[0]
        sub_widget(gate_set, v, g)
        for i in range(n - 1):
            X & g[i]
        sub_widget(gate_set, v, g)
        for i in range(n):
            CX & [g[n - 1], v[i]]


def c_incrementer(gate_set, control, v, g_aug):
    """
    1-controlled incremente v by 1, with borrowed qubits g.

    Constructed by attaching the control qubit to the little-end of v,
    and apply an (n+1)-bit incrementer() to it.

    Args:
        control(int): index of 1 qubit.
        v(Qureg): indices of n qubits.
        g(Qureg): indices of n + 1 qubits(more qubits are OK).
    """
    n = len(v)
    m = len(g_aug)
    if m < n + 1:
        print("no edequate ancilla bits")
    g = g_aug[0:n + 1]
    vc = v + [control]
    with gate_set:
        incrementer(gate_set, vc, g)
        X & vc[n]


def adder_rec(gate_set, control, x, c_bitwise, ancilla, ancilla_g):
    """
    The recursively applied partial-circuit in adder().

    Controlled version constructed by changing carry() to c_carry().

    Args:
        control(list): indices of 0 or 1 qubit.
        x(Qureg): indices of n qubits.
        ancilla(Qubit): index of 1 qubit.
        ancilla_g(Qubit): index of 1 qubit,
            might be used as borrowed qubit in c_incrementer
            when x_H and x_L are of the same length.
        c_bitwise(char array): n bits '0'-'1' array, representing binary int c.
    """
    n = len(x)
    # n_control = len(control)
    if n == 1:
        return
    mid = n // 2
    x_H = x[0:mid]
    x_L = x[mid:n]
    c_H = c_bitwise[0:mid]
    c_L = c_bitwise[mid:n]
    g = x_L + [ancilla_g]
    with gate_set:
        c_incrementer(gate_set, ancilla, x_H, g)
        for i in range(mid):
            CX & [ancilla, x_H[i]]
        carry(gate_set, control, x_L, c_L, x_H, ancilla)
        c_incrementer(gate_set, ancilla, x_H, g)
        carry(gate_set, control, x_L, c_L, x_H, ancilla)
        for i in range(mid):
            CX & [ancilla, x_H[i]]
        adder_rec(gate_set, control, x_L, c_L, ancilla, ancilla_g)
        adder_rec(gate_set, control, x_H, c_H, ancilla, ancilla_g)


def adder(gate_set, control, x, c, ancilla, ancilla_g):
    """
    Compute x(quantum) + c(classical) with borrowed qubits, at most 1-controlled.

    Args:
        control(list): indices of 0 or 1 qubit.
        x(list): indices of n qubits.
        ancilla(int): index of 1 qubit, borrowed ancilla.
        ancilla_g(int): index of 1 qubit, borrowed ancilla.
        c(int): integer.
    """
    n = len(x)
    c_bitwise = int2bitwise(c, n)
    with gate_set:
        adder_rec(gate_set, control, x, c_bitwise, ancilla, ancilla_g)
        for i in range(n):
            if c_bitwise[i] == "1":
                var_controlled_X(gate_set, control, x[i])


def c_sub(gate_set, control, x, c, ancilla, ancilla_g):
    """
    Compute x(quantum) - c(classical) with borrowed qubits, 1-controlled.

    Constructed on the basis of c_adder() with complement technique.

    Args:
        control(Qubit): 1 qubit.
        x(Qureg): n qubits.
        ancilla(Qubit): 1 qubit, borrowed ancilla.
        ancilla_g(Qubit): 1 qubit, borrowed ancilla.
        c(int): integer.
    """
    n = len(x)
    c_complement = 2 ** n - c
    cc_bitwise = int2bitwise(c_complement, n)
    with gate_set:
        adder_rec(gate_set, [control], x, cc_bitwise, ancilla, ancilla_g)
        for i in range(n):
            if cc_bitwise[i] == "1":
                CX & [control, x[i]]


def compare(gate_set, control, b, c, g_aug, indicator):
    """
    compare b and c with borrowed qubits g_aug.
    The Indicator toggles if c > b, not if c <= b.
    at most 2-controlled.

    Constructed on the basis of carry().

    Args:
        control(list): 0 or 2 qubit.
        b(Qureg): n qubits.
        g_aug(Qureg): n-1 qubits(more qubits are OK).
        indicator(Qubit): 1 qubit.
        c(int): integer with less-than-n length.
    """
    n = len(b)
    m = len(g_aug)
    if m < n - 1:
        print("No edequate ancilla bits when compare\n")
        return
    c_bitwise = int2bitwise(c, n)
    with gate_set:
        for i in range(n):
            X & b[i]
        carry(gate_set, control, b, c_bitwise, g_aug, indicator)
        for i in range(n):
            X & b[i]


def adder_mod(gate_set, control, b, a, N, g, indicator):
    """
    Compute b(quantum) + a(classical) mod N(classical),
    with borrowed qubits g and ancilla qubit indicator, at most 2-controlled.

    Args：
        control(list): 0 or 2 qubit.
        b(Qreg): n qubits.
        g(Qureg): n-1 borrowed qubits(more qubits are OK).
        indicator(Qubit): 1 ancilla qubit.
        a(int): integer less than N.
        N(int): integer.

    Note that this circuit works only when n > 2.
    So for smaller numbers, use another design.
    """
    if len(b) <= 2:
        raise Exception("The numbers should be more than 2-length to use HRS circuits.")
    with gate_set:
        compare(gate_set, control, b, N - a, g, indicator)
        adder(gate_set, [indicator], b, a, g[0], g[1])
        var_controlled_X(gate_set, control, indicator)
        c_sub(gate_set, indicator, b, N - a, g[0], g[1])
        var_controlled_X(gate_set, control, indicator)
        compare(gate_set, control, b, a, g, indicator)
        var_controlled_X(gate_set, control, indicator)


def adder_mod_reversed(gate_set, control, b, a, N, g, indicator):
    """
    The reversed circuit of adder_mod()
    """
    adder_mod(gate_set, control, b, N - a, N, g, indicator)


def mul_mod_raw(gate_set, control, x, a, b, N, indicator):
    """
    Compute b(quantum) + x(quantum) * a(classical) mod N(classical),
    with target qubits b and ancilla qubit indicator.

    Args:
        control(Qubit): 0 or 1 qubit.
        x(Qureg): n qubits.
        b(Qureg): n qubits, target.
        indicator(Qubit): 1 ancilla qubit.
        a(int): integer.
        N(int): integer.

    Note that this circuit works only when n > 2.
    So for smaller numbers, use another design.
    """

    n = len(x)
    if n <= 2:
        raise Exception("The numbers should be more than 2-length to use HRS circuits.")
    a_list = []
    for i in range(n):
        a_list.append(a)
        a = (a * 2) % N
    with gate_set:
        for i in range(n):
            # borrow all the n-1 unused qubits in x
            g = x[:n - i - 1] + x[n - i:]
            if len(control) == 0:
                adder_mod(gate_set, [x[n - 1 - i]], b, a_list[i], N, g, indicator)
            elif len(control) == 1:
                adder_mod(
                    gate_set, control + [x[n - 1 - i]], b, a_list[i], N, g, indicator
                )


def mul_mod_raw_reversed(
    gate_set, control, x, a, b, N, indicator
):  # TODO: this can be done classically with `mul_mod_raw`
    """
    The reversed circuit of mul_mod_raw()
    Note that this circuit works only when n > 2.
    So for smaller numbers, use another design.
    """

    n = len(x)
    if n <= 2:
        raise Exception("The numbers should be more than 2-length to use HRS circuits.")
    a_list = []
    for i in range(n):
        a_list.append(a)
        a = a * 2 % N
    with gate_set:
        for i in range(n):
            g = x[:i] + x[i + 1:]
            if len(control) == 0:
                adder_mod(gate_set, [x[i]], b, N - a_list[n - i - 1], N, g, indicator)
            elif len(control) == 1:
                adder_mod(
                    gate_set,
                    control + [x[i]],
                    b,
                    N - a_list[n - i - 1],
                    N,
                    g,
                    indicator,
                )
            else:
                raise ValueError()


def mul_mod(gate_set, control, x, a, ancilla, N, indicator):
    """
    Compute x(quantum) * a(classical) mod N(classical), with ancilla qubits.

    Args:
        control(Qubit): 0 or 1 qubit.
        x(Qureg): n qubits.
        ancilla(Qureg): n qubits.
        indicator(Qubit): 1 qubit.
        a(int): integer.
        N(int): integer.

    Note that this circuit works only when n > 2.
    So for smaller numbers, use another design.
    """

    n = len(x)
    if n <= 2:
        raise Exception("The numbers should be more than 2-length to use HRS circuits.")
    a_r = mod_reverse(a, N)
    with gate_set:
        mul_mod_raw(gate_set, control, x, a, ancilla, N, indicator)
        # Swap
        for i in range(n):
            if len(control) == 0:
                Swap & [x[i], ancilla[i]]
            elif len(control) == 1:
                CSwap & [control[0], x[i], ancilla[i]]
            else:
                raise ValueError()
        mul_mod_raw_reversed(gate_set, control, x, a_r, ancilla, N, indicator)


class HRSAdder(object):
    @staticmethod
    def execute(n, c):
        """
        Compute x(quantum) + c(classical) with borrowed qubits.

        Args:
            n(int): length of numbers
            c(int): the constant added to the quantum number

        Quregs:
            x(Qureg): n qubits.
            ancilla(Qubit): 1 qubit, borrowed ancilla.
            ancilla_g(Qubit): 1 qubit, borrowed ancilla.
        """

        gate_set = CompositeGate()
        qubit_x = list(range(n))
        ancilla = n
        ancilla_g = n + 1

        adder(gate_set, [], qubit_x, c, ancilla, ancilla_g)

        return gate_set


class HRSAdderMod(object):
    @staticmethod
    def execute(n, a, N):
        """
        Compute b(quantum) + a(classical) mod N(classical),
        with borrowed qubits g and ancilla qubit indicator.

        Args:
            n(int): length of numbers
            a(int): the constant added to the quantum number
            N(int): the modulus

        Quregs：
            b(Qreg): n qubits.
            g(Qureg): n-1 borrowed qubits(more qubits are OK).
            indicator(Qubit): 1 ancilla qubit.

        Note that this circuit works only when n > 2.
        So for smaller numbers we use another design.
        """

        if n <= 2:
            raise Exception(
                "The numbers should be more than 2-length to use HRS circuits."
            )

        gate_set = CompositeGate()
        qubit_b = list(range(n))
        g = list(range(n, 2 * n - 1))
        indicator = 2 * n - 1

        adder_mod(gate_set, [], qubit_b, a, N, g, indicator)

        return gate_set


class HRSMulMod(object):
    @staticmethod
    def execute(n, a, N):
        """
        Compute x(quantum) * a(classical) mod N(classical), with ancilla qubits.

        Args:
            n(int): length of numbers
            a(int): the constant multiplied to the quantum number
            N(int): the modulus

        Quregs:
            x(Qureg): n qubits.
            ancilla(Qureg): n qubits.
            indicator(Qubit): 1 qubit.

        Note that this circuit works only when n > 2.
        So for smaller numbers we use another design.
        """

        if n <= 2:
            raise Exception(
                "The numbers should be more than 2-length to use HRS circuits."
            )

        gate_set = CompositeGate()
        qubit_x = list(range(n))
        ancilla = list(range(n, 2 * n))
        indicator = 2 * n

        mul_mod(gate_set, [], qubit_x, a, ancilla, N, indicator)

        return gate_set


class CHRSMulMod(object):
    @staticmethod
    def execute(n, a, N):
        """
        Compute x(quantum) * a(classical) mod N(classical), with ancilla qubits, 1-controlled.

        Args:
            n(int): length of numbers
            a(int): the constant multiplied to the quantum number
            N(int): the modulus

        Quregs:
            control(Qubit): 1 qubit.
            x(Qureg): n qubits.
            ancilla(Qureg): n qubits.
            indicator(Qubit): 1 qubit.

        Note that this circuit works only when n > 2.
        So for smaller numbers we use another design.
        """
        if n <= 2:
            raise Exception(
                "The numbers should be more than 2-length to use HRS circuits."
            )

        gate_set = CompositeGate()
        qubit_x = list(range(n))
        ancilla = list(range(n, 2 * n))
        indicator = 2 * n
        control = 2 * n + 1

        mul_mod(gate_set, [control], qubit_x, a, ancilla, N, indicator)

        return gate_set



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/3/20 4:52 下午
# @Author  : Han Yu
# @File    : __init__.py

from .google_set import GoogleSet
from .ibmq_set import IBMQSet
from .ionq_set import IonQSet
from .ustc_set import USTCSet
from .nam_set import NamSet
from .origin_set import OriginSet



from typing import Union, List


class Operator:
    """ The SuperClass of all the operator. """
    @property
    def targets(self):
        return self._targets

    @property
    def targs(self):
        return self._targs

    @property
    def cargs(self):
        return []

    @targs.setter
    def targs(self, targets: Union[List, int]):
        assert isinstance(targets, (int, list))
        if isinstance(targets, int):
            self._targs = [targets]
        else:
            self._targs = targets

    @property
    def targ(self):
        return self._targs[0]

    def __init__(
        self,
        targets: int
    ):
        """
        Args:
            targets (int): the number of the target bits of the gate
        """
        assert targets >= 0 and isinstance(targets, int), f"targets must be a positive integer, not {type(targets)}"
        self._targets = targets
        self._targs = []

    def __and__(self, targets: Union[List[int], int],):
        """ Assigned the trigger's target qubits.

        Args:
            targets (Union[List[int], int]): The indexes of target qubits.
        """
        if isinstance(targets, int):
            assert targets >= 0 and self.targets == 1
            self._targs = [targets]
        elif isinstance(targets, list):
            assert len(targets) == self.targets
            for q in targets:
                assert q >= 0 and isinstance(q, int), f"targets must be a positive integer, not {q}"

            self._targs = targets
        else:
            raise TypeError(f"qubits must be one of [List[int], int], not {type(targets)}")

        return self

    def __or__(self, targets):
        """ Append the trigger to the circuit.

        Args:
            targets (Circuit): The circuit which contains the trigger.
        """
        try:
            targets.append(self)
        except Exception as e:
            raise TypeError(f"Failure to append Trigger to targets, due to {e}")



import subprocess
from kubenerte import client

from server.utils.get_config import get_default_job_config


class KubeController:
    def __init__(self):
        self._api_instance = client.BatchV1Api()

    @staticmethod
    def generate_kube_job_yaml(job_info: dict) -> dict:
        """ Get kubernetes job yaml by given job information

        Return:
        job_yaml(dict) = {
            apiVersion: batch/v1
            kind: Job
            metadata:
                name: jobname
                namespace: username
            spec:
                completions: 1
                parallelism: 1
                ttlSecondsAfterFinished: 10
                backoffLimit: 0
                template:
                    spec:
                        restartPolicy: Never
                        containers:
                          - name: job_type
                            image: quict:v1.0
                            imagePullPolicy: IfNotPresent
                            command: ['python', job_type.py]
                            env:
                                - name: circuit
                                  value: "/data/circuit.qasm"
                                - name: layout
                                  value: "/data/layout.json"
                                - other option arguments
                            resources:
                                requests:
                                    cpu: 0.1
                                    memory: 32Mi
                                limits:
                                    cpu: 0.5
                                    memory: 32Mi
                            volumeMounts:
                                - name: origincircuit
                                mountPath: /data
                                readOnly: False
                        volumes:
                        - name: origincircuit
                            hostPath:
                            path: /data/UserName/JobName
            }
        """
        return get_default_job_config(job_info)

    def start_job(self, job_info):
        """ kubectl apply -f path"""
        # Standardize start job deployment.
        k8s_job = self.generate_kube_job_yaml(job_info)

        # Create and apply k8s config
        self._api_instance.create_namespaced_job(body=k8s_job, namespace=job_info['username'])

    def delete_job(self, job_name, user_name):
        """ Delete kubernate's Job

        Args:
            job_name (_type_): _description_
            user_name (_type_): _description_
        """
        try:
            self._api_instance.delete_namespaced_job(
                name=job_name,
                namespace=user_name,
                body=client.V1DeleteOptions(
                    propagation_policy='Foreground',
                    grace_period_seconds=5
                )
            )
        except Exception as e:
            raise e

    def get_job_status(self, job_name: str, username: str):
        api_response = self._api_instance.read_namespaced_job_status(
            name=job_name,
            namespace=username
        )

        return api_response.status

    def get_job_info(self, job_name, username):
        """ Return the job's running information

        Args:
            job_name (_type_): _description_
            username (_type_): _description_
        """
        # Get resources
        job_list = self._api_instance.list_namespaced_job(
            namespace=username, pod_name=job_name
        ).to_dict()["items"]

        return job_list

    def list_jobs(self, username):
        job_list = self._api_instance.list_namespaced_job(namespace=username).to_dict()["items"]

        return job_list



from collections import defaultdict
import numpy as np

from QuICT.core.gate import Unitary
from QuICT.ops.linalg.cpu_calculator import multiply, dot, tensor, MatrixPermutation


class Optimizer:
    """ The quantum circuit optimizer, used to optimize the quantum circuit
    before simulation. Currently, it only working with the 1-, 2-qubits quantum
    gates.

    The rules:
        1. For the consecutive 1-qubit gates, it will be combined into a new 1-qubit Unitary
        gate.
        2. Merge the 1-qubit gates into the connected 2-qubits gate, if they have same qubit.
        3. If any 2-qubits gate has same target and control qubit indexes, and no other gates
        between them, merge its into a new 2-qubits Unitary gate.
    """
    def __init__(self):
        self._opt_gates = []    # The gates list after optimizer
        self._gates_by_qubit = defaultdict(list)    # Temporarily store the 1-qubit gates by qubits
        self._two_qubits_opt_gates_idxes_dict = {}  # Record the qubit indexes of the 2-qubits gates
        self._based_tensor_matrix = np.identity(2, dtype=np.complex128)

    def optimize(self, gates: list) -> list:
        """ Depending on the optimized rules, generate the optimized gates by the given quantum gates.

        Args:
            gates (list[Gate]): The list of gates before optimization

        Returns:
            list: The list of the optimized gates
        """
        self._cache_clear()

        for gate in gates:
            # The qubit indexes of the gate
            qubit_idxes = gate.cargs + gate.targs

            if gate.is_special():   # Do not optimized any special quantum gates
                self._1qubit_gates_clean(qubit_idxes)
                self._opt_gates.append(gate)
                continue

            if len(qubit_idxes) == 1:       # 1-qubit gate
                self._gates_by_qubit[gate.targ].append(gate)
            elif len(qubit_idxes) == 2:     # 2-qubits gate
                self.two_qubits_gates_combined(
                    gate
                )

                for idx in qubit_idxes:
                    del self._gates_by_qubit[idx]
            else:   # TODO: not support the gates with more than 2 qubits
                self._1qubit_gates_clean(qubit_idxes)
                self._opt_gates.append(gate)
                continue

        # Merge the rest 1-qubit gates into a Unitary gate.
        for qubit, gates in self._gates_by_qubit.items():
            if gates:
                self.single_gates_combined(qubit)

        return self._opt_gates

    def _cache_clear(self):
        """ Initial the Optimizer """
        self._opt_gates = []
        self._gates_by_qubit = defaultdict(list)
        self._two_qubits_opt_gates_idxes_dict = {}

    def _is_diagonal_matrix(self, matrix):
        return np.allclose(np.diag(np.diag(matrix)), matrix)

    def _1qubit_gates_clean(self, args):
        for arg in args:
            if self._gates_by_qubit[arg]:
                self.single_gates_combined(arg)
                del self._gates_by_qubit[arg]

    def _gate_combined(self, qubit_idxes: list):
        assert len(qubit_idxes) == 2
        cidx_matrix = self.single_gates_combined(qubit_idxes[0], matrix_only=True)
        tidx_matrix = self.single_gates_combined(qubit_idxes[1], matrix_only=True)

        return tensor(cidx_matrix, tidx_matrix)

    def single_gates_combined(self, qubit, matrix_only: bool = False):
        """ The optimized algorithm for 1-qubit gates with same qubit.

        Args:
            gates ([Gate]): The 1-qubit gates in the same qubit.
            matrix_only (bool, optional): return the Unitary gate or matrix. Defaults to False.

        Returns:
            (UnitaryGate, np.array): the new Unitary gate or the gate matrix
        """
        gates = self._gates_by_qubit[qubit]
        if not gates:
            return self._based_tensor_matrix

        if len(gates) == 1 and not matrix_only:
            self._opt_gates.append(gates[0])
            return

        based_matrix = gates[0].matrix
        for gate in gates[1:]:
            if self._is_diagonal_matrix(based_matrix) and gate.is_diagonal():  # Using multiply for diagonal gates
                based_matrix = multiply(gate.matrix, based_matrix)
            else:   # Using dot for non-diagonal gates
                based_matrix = dot(gate.matrix, based_matrix)

        if matrix_only:
            return based_matrix

        # Add the optimized quantum gate
        opt_gate = Unitary(based_matrix)
        opt_gate.targs = qubit
        self._opt_gates.append(opt_gate)

    def two_qubits_gates_combined(self, two_qubit_gate):
        """ The optimized algorithm for 2-qubits gates and related 1-qubit gates.

        Args:
            two_qubit_gate (Gate): The 2-qubits quantum gate
            cidx_gates (gates, None): The 1-qubit quantum gates in the control index
            tidx_gates (gates, None): The 1-qubit quantum gates in the target index
            gate_idx_bit (int): The bit-represent for the indexes of the 2-qubits quantum gate
        """
        qubit_idxes = two_qubit_gate.cargs + two_qubit_gate.targs
        gate_idx_bit = np.sum([1 << qidx for qidx in qubit_idxes])
        reverse = qubit_idxes[0] > qubit_idxes[1]

        # Combined the gate matrix of the control and target indexes into the 2-qubits gate matrix
        single_gates_combined_matrix = self._gate_combined(qubit_idxes)
        if self._is_diagonal_matrix(single_gates_combined_matrix) and two_qubit_gate.is_diagonal():
            opt_gate_matrix = multiply(two_qubit_gate.matrix, single_gates_combined_matrix)
        else:
            opt_gate_matrix = dot(two_qubit_gate.matrix, single_gates_combined_matrix)

        if reverse:
            MatrixPermutation(opt_gate_matrix, np.array([1, 0]), changeInput=True)

        # Combined the two-qubits gates
        if gate_idx_bit in self._two_qubits_opt_gates_idxes_dict.keys():
            pre_gate = self._opt_gates[self._two_qubits_opt_gates_idxes_dict[gate_idx_bit]]
            if pre_gate.is_diagonal() and self._is_diagonal_matrix(opt_gate_matrix):
                pre_gate.matrix = multiply(opt_gate_matrix, pre_gate.matrix)
            else:
                pre_gate.matrix = dot(opt_gate_matrix, pre_gate.matrix)
        else:
            overwritten_idx = \
                [idx_bit for idx_bit in self._two_qubits_opt_gates_idxes_dict.keys() if idx_bit & gate_idx_bit]
            for idx in overwritten_idx:
                del self._two_qubits_opt_gates_idxes_dict[idx]

            opt_gate = Unitary(opt_gate_matrix)
            opt_gate.targs = qubit_idxes

            self._two_qubits_opt_gates_idxes_dict[gate_idx_bit] = len(self._opt_gates)
            self._opt_gates.append(opt_gate)



from abc import abstractmethod


class RemoteSimulator(object):
    """ The based class for remote simulator, currently support Qiskit and QCompute.

    Args:
        backend ([str]): The backend for the remote simulator.
        shots ([int]): The repeat times of running circuit, always be a positive integer.
    """
    def __init__(self, backend, shots):
        assert shots >= 1, "The shots must be a positive integer."
        self._backend = backend
        self._shots = shots

    @abstractmethod
    def run(self):
        pass

    def get_backend(self):
        return self.__BACKEND



import ctypes
import threading


class TimeoutThread(threading.Thread):
    def __init__(self, func, *args, **kwargs):
        threading.Thread.__init__(self)
        self.result = None
        self.exception = None
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def run(self):
        try:
            self.result = self.func(*self.args, **self.kwargs)
        except Exception as e:
            self.exception = e


def _async_raise(tid, excobj):
    tid = ctypes.c_long(tid)
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(excobj))
    if res == 0:
        raise ValueError("nonexistent thread id")
    elif res > 1:
        ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, 0)
        raise SystemExit("PyThreadState_SetAsyncExc failed")


def terminal_thread(thread):
    assert thread.is_alive(), "thread must be started"
    _async_raise(thread.ident, SystemExit)


def timeout():
    def decorator(func):
        def wraps(self, *args, **kwargs):
            t_out = self._timeout
            new_thread = TimeoutThread(func, self, *args, **kwargs)
            new_thread.start()
            new_thread.join(t_out)

            if new_thread.is_alive():
                terminal_thread(new_thread)
                raise TimeoutError("Running time exceed the time limit.")
            elif new_thread.exception:
                raise new_thread.exception
            else:
                return new_thread.result

        return wraps

    return decorator



from ..data_def import ReplayMemory, TrainConfig, Transition
from ..net.rl_agent import Agent


class Actor:
    def __init__(self, rank: int, config: TrainConfig):
        self.rank = rank
        self.config = config
        self.agent = Agent(config=config)
        self.policy_net = None
        self.replay = ReplayMemory(capacity=config.memory_sync_period)

    def explore(self):
        action = self.agent.select_action(
            policy_net=self.policy_net,
            policy_net_device=self.config.device,
            epsilon_random=True,
        )

        prev_state, next_state, reward, terminated = self.agent.take_action(
            action=action
        )

        if terminated:
            next_state = None
            self.agent.reset_explore_state()

        self.replay.push(
            Transition(
                state=prev_state,
                action=action,
                next_state=next_state,
                reward=reward,
            )
        )

        return reward



import os
import subprocess


command_file_path = os.path.join(
    os.path.dirname(__file__),
    "../script/benchmark.py"
)


def benchmark(gpu: bool):
    _ = subprocess.call(
        f"python {command_file_path} {gpu}", shell=True
    )



import os
import shutil
from typing import Union, List

from QuICT.core import Circuit
from QuICT.core.gate import GateType
from QuICT.lib.qasm.exceptions import QasmError
from QuICT.tools import Logger
from QuICT.tools.interface import OPENQASMInterface
from .circuit_lib_sql import CircuitLibDB
from .get_benchmark_circuit import BenchmarkCircuitBuilder


logger = Logger("QuICT_Circuit_Library")


class CircuitLib:
    """
    A class handling QuICT circuit library.

    Args:
        output_type (str, optional): one of [circuit, qasm, file]. Defaults to "circuit".
        output_path (str, optional): The path to store qasm file if output type is file.
            Defaults to the current working path.
    """

    __DEFAULT_TYPE = ["template", "random", "algorithm", "benchmark", "instructionset"]
    __DEFAULT_CLASSIFY = {
        "random": [
            "aspen-4", "ourense", "rochester", "sycamore", "tokyo",
            "ctrl_unitary", "diag", "single_bit", "ctrl_diag",
            "google", "ibmq", "ionq", "ustc", "nam", "origin"
        ],
        "algorithm": ["adder", "clifford", "cnf", "grover", "maxcut", "qft", "qnn", "quantum_walk", "vqe"],
        "benchmark": ["highly_entangled", "highly_parallelized", "highly_serialized", "mediate_measure"],
    }
    __DEFAULT_GATESET_for_RANDOM = {
        "google": [[GateType.fsim], [GateType.sx, GateType.sy, GateType.sw, GateType.rx, GateType.ry]],
        "ibmq": [[GateType.cx], [GateType.rz, GateType.sx, GateType.x]],
        "ionq": [[GateType.rxx], [GateType.rx, GateType.ry, GateType.rz]],
        "ustc": [[GateType.cx], [GateType.rx, GateType.ry, GateType.rz, GateType.h, GateType.x]],
        "nam": [[GateType.cx], [GateType.h, GateType.rz]],
        "origin": [[GateType.cx], [GateType.u3]],
        "ctrl_diag": [[GateType.crz, GateType.cu1, GateType.cz], []],
        "ctrl_unitary": [[GateType.cx, GateType.cy, GateType.ch, GateType.cu3], []],
        "diag": [[], [
            GateType.t, GateType.rz, GateType.z, GateType.sdg, GateType.tdg, GateType.u1, GateType.s, GateType.id
        ]],
        "single_bit": [[], [
            GateType.x, GateType.y, GateType.z, GateType.u1, GateType.u2, GateType.u3, GateType.tdg,
            GateType.sdg, GateType.h, GateType.s, GateType.t, GateType.rx, GateType.ry, GateType.rz
        ]],
        "unitary": [[GateType.swap], [GateType.y, GateType.rx, GateType.ry, GateType.u2, GateType.u3]]
    }
    __LIB_PATH = os.path.join(
        os.path.dirname(__file__),
        '../../lib/circuitlib'
    )

    def __init__(
        self,
        output_type: str = "circuit",
        output_path: str = '.'
    ):
        self._output_type = output_type
        self._output_path = output_path
        logger.debug(f"Initial Circuit Library with output {self._output_type}.")

        self._db = CircuitLibDB()
        if output_type == "file":
            if not os.path.isdir(output_path):
                os.makedirs(output_path)

    @property
    def size(self) -> int:
        return self._db.size()

    def _get_circuit_from_qasm(self, file_path: str) -> Circuit:
        """ Load Circuit from a qasm file. """
        qasm = OPENQASMInterface.load_file(file_path)
        if not qasm.valid_circuit:
            logger.warn("Failure to load circuit from qasm.")
            raise QasmError("Missing input file")

        split_path = file_path.split('/')
        cir_name, cir_classify, cir_type  = split_path[-1:-4:-1]
        cir_name = os.path.splitext(cir_name)[0]
        qasm.circuit.name = "+".join([cir_type, cir_classify, cir_name])
        return qasm.circuit

    def _get_string_from_qasm(self, file_path: str) -> str:
        """ Return qasm string from qasm file. """
        return open(file_path, 'r').read()

    def _copy_qasm_file(self, file_path: str):
        """ Copy target qasm file to the given output path. """
        shutil.copy(file_path, self._output_path)

    def _get_all(self, folder: str, files: list) -> Union[List, None]:
        """ Load all qasm files in the list of files.

        Args:
            folder (str): The path of the qasm folder
            files (list): the list of file names

        Returns:
            List[(Circuit|string)]: _description_
        """
        circuit_all = []
        for file in files:
            file_path = os.path.join(folder, file[0])
            if self._output_type == "circuit":
                circuit_all.append(self._get_circuit_from_qasm(file_path))
            elif self._output_type == "qasm":
                circuit_all.append(self._get_string_from_qasm(file_path))
            else:
                self._copy_qasm_file(file_path)

        return circuit_all

    def _get_circuit_from_benchmark(self, classify, width, size):
        if classify == "highly_entangled":
            circuit = BenchmarkCircuitBuilder.entangled_circuit_build(width, size)
        elif classify == "highly_parallelized":
            circuit = BenchmarkCircuitBuilder.parallelized_circuit_build(width, size)
        elif classify == "highly_serialized":
            circuit = BenchmarkCircuitBuilder.serialized_circuit_build(width, size)
        else:
            circuit = BenchmarkCircuitBuilder.mediate_measure_circuit_build(width, size)

        return circuit

    def _get_all_from_generator(
        self,
        type: str,
        classify: str,
        gateset: list,
        prob: list,
        qubits_interval: Union[list, int],
        max_size: int = None,
        max_depth: int = None
    ):
        if isinstance(qubits_interval, int):
            qubits_interval = list(range(2, qubits_interval + 1))

        circuit_list = []
        size_interval = [3, 5, 10, 20]
        for width in qubits_interval:
            for size in size_interval:
                if max_size is not None and size * width > max_size:
                    continue

                if type == "random":
                    circuit = Circuit(width)
                    circuit.random_append(size * width, gateset, True, prob)
                    depth = circuit.depth()
                else:
                    circuit = self._get_circuit_from_benchmark(classify, width, size * width)
                    depth = circuit.depth()

                if max_depth is None or depth <= max_depth:
                    circuit.name = "+".join([type, classify, f"w{width}_s{size * width}_d{depth}"])
                    circuit_list.append(circuit)

        if self._output_type == "circuit":
            return circuit_list
        elif self._output_type == "qasm":
            return [circuit.qasm() for circuit in circuit_list]
        else:
            for circuit in circuit_list:
                output_file = os.path.join(self._output_path, f"{circuit.name}.qasm")
                circuit.qasm(output_file)

    def get_template_circuit(
        self,
        qubits_interval: Union[list, int] = None,
        max_size: int = None,
        max_depth: int = None,
        typelist: list = None
    ) -> Union[List[Union[Circuit, str]], None]:
        """
        Get template circuits in QuICT circuit library. A template will be loaded if
        it satisfies the following restrictions:
            1. the circuit in the given classify.
            2. its number of qubits <= `max_width`,
            3. its number of gates <= `max_size`,
            4. its depth <= `max_depth`.
            5. its gates' types are in `typelist`

        Restrictions will be ignored if not specified.

        Args:
            qubits_interval (Union[List, int], optional): The interval of qubit number, if it givens an interger,
                it equals to the interval of [2, qubits_interval]. The qubits' number range is [1, 5].
            max_size(int): max number of gates, range is [2, 6].
            max_depth(int): max depth of circuit, range is [2, 9].
            typelist(Iterable[GateType]): list of allowed gate types

        Returns:
            (List[Circuit | String] | None): Return the list of output circuit order by output_type.
        """
        path = os.path.join(self.__LIB_PATH, "template")
        files = self._db.circuit_filter("template", "template", qubits_interval, max_size, max_depth)

        ret = self._get_all(path, files)

        if typelist is not None:
            filtered = []
            for each in ret:
                if all([g.type in typelist for g in each.gates]):
                    filtered.append(each)
            ret = filtered

        return ret

    def get_random_circuit(
        self,
        classify: str,
        qubits_interval: Union[list, int],
        max_size: int = None,
        max_depth: int = None
    ) -> Union[List[Union[Circuit, str]], None]:
        """Get random circuits in QuICT circuit library. A template will be loaded if
        it satisfies the following restrictions:
            1. the circuit in the given classify.
            2. its number of qubits <= `max_width`,
            3. its number of gates <= `max_size`,
            4. its depth <= `max_depth`.

        Restrictions will be ignored if not specified.

        Args:
            classify (str): one of ["aspen-4", "ourense", "rochester", "sycamore", "tokyo", \
                "ctrl_unitary", "diag", "single_bits", "ctrl_diag", "google", "ibmq", "ionq", "ustc", "nam", "origin"]
            qubits_interval (Union[List, int], optional): The interval of qubit number, if it givens an interger,
                it equals to the interval of [2, qubits_interval].
            max_size(int): max number of gates.
            max_depth(int): max depth of circuit.

        Returns:
            (List[Circuit | String] | None): Return the list of output circuit order by output_type.
        """
        assert classify in self.__DEFAULT_CLASSIFY['random'], "error classify."
        if classify in self.__DEFAULT_CLASSIFY['random'][:5]:
            path = os.path.join(self.__LIB_PATH, 'random', classify)
            files = self._db.circuit_filter("random", classify, qubits_interval, max_size, max_depth)

            return self._get_all(path, files)
        else:   # Generate random circuit with given limitation
            gate_2q, gate_1q = self.__DEFAULT_GATESET_for_RANDOM[classify]
            if len(gate_2q) > 0 and len(gate_1q) > 0:
                prob = [0.2 / len(gate_2q)] * len(gate_2q) + [0.8 / len(gate_1q)] * len(gate_1q)
            else:
                prob = None

            return self._get_all_from_generator(
                "random", classify, gate_2q + gate_1q, prob, qubits_interval, max_size, max_depth
            )

    def get_algorithm_circuit(
        self,
        classify: str,
        qubits_interval: Union[list, slice] = None,
        max_size: int = None,
        max_depth: int = None
    ) -> Union[List[Union[Circuit, str]], None]:
        """ Get algorithm circuits in QuICT circuit library. A template will be loaded if
        it satisfies the following restrictions:
            1. the circuit in the given classify.
            2. its number of qubits <= `max_width`,
            3. its number of gates <= `max_size`,
            4. its depth <= `max_depth`.

        Restrictions will be ignored if not specified.

        Args:
            classify (str): one of ["adder", "clifford", "grover", "qft", "vqe"]
            qubits_interval (Union[List, int], optional): The interval of qubit number, if it givens an interger,
                it equals to the interval of [2, qubits_interval].
            max_size(int): max number of gates.
            max_depth(int): max depth of circuit.

        Returns:
            (List[Circuit | String] | None): Return the list of output circuit order by output_type.
        """
        assert classify in self.__DEFAULT_CLASSIFY['algorithm'], "error classify."
        path = os.path.join(self.__LIB_PATH, 'algorithm', classify)
        files = self._db.circuit_filter("algorithm", classify, qubits_interval, max_size, max_depth)

        return self._get_all(path, files)

    def get_benchmark_circuit(
        self,
        classify: str,
        qubits_interval: Union[list, int] = None,
        max_size: int = None,
        max_depth: int = None
    ) -> Union[List[Union[Circuit, str]], None]:
        """ Get experiment circuits in QuICT circuit library. A template will be loaded if
        it satisfies the following restrictions:
            1. the circuit in the given classify.
            2. its number of qubits <= `max_width`,
            3. its number of gates <= `max_size`,
            4. its depth <= `max_depth`.

        Restrictions will be ignored if not specified.

        Args:
            classify (str): one of ["highly_entangled", "highly_parallelized", "highly_serialized", "mediate_measure"]
            qubits_interval (Union[List, int], optional): The interval of qubit number, if it givens an interger,
                it equals to the interval of [2, qubits_interval].
            max_size(int): max number of gates.
            max_depth(int): max depth of circuit.

        Returns:
            (List[Circuit | String] | None): Return the list of output circuit order by output_type.
        """
        assert classify in self.__DEFAULT_CLASSIFY['benchmark'], "error experiment classify."
        # Generate benchmark circuit with given limitation
        return self._get_all_from_generator("benchmark", classify, [], [], qubits_interval, max_size, max_depth)

    def get_circuit(
        self,
        type: str,
        classify: str = "template",
        qubits_interval: Union[list, int] = None,
        max_size: int = None,
        max_depth: int = None
    ) -> Union[List[Union[Circuit, str]], None]:
        """Get the target circuits from QuICT Circuit Library.

        Args:
            type (str): The type of circuits, one of [template, random, algorithm, benchmark].
            classify (str, optional): The classify of selected circuit's type.
                For template circuit's type, classify must be template;
                For random circuit's type, classify is one of
                    [aspen-4, ourense, rochester, sycamore, tokyo, ctrl_unitary, diag, single_bit, ctrl_diag,
                     google, ibmq, ionq, ustc, nam, origin]
                For algorithm circuit's type, classify is one of
                    [adder, clifford, grover, qft, vqe, cnf, maxcut]
                For benchmark circuit's type, classify is one of
                    [highly_entangled, highly_parallelized, highly_serialized, mediate_measure]

            qubits_interval (Union[List, int], optional): The interval of qubit number, if it givens an interger,
                it equals to the interval of [1, qubits_interval].
            max_size (int, optional): upper bound of circuit size. If None, no limitation on size, default to None.
            max_depth (int, optional): upper bound of circuit depth. If None, no limitation on depth, default to None.

            WARNING: qubits_interval need to be assigned for getting random circuit and benchmark circuit.

        Returns:
            (List[Circuit | String] | None): Return the list of output circuit order by output_type.
        """
        if type not in self.__DEFAULT_TYPE:
            raise KeyError("error_type")

        if type != "template" and classify not in self.__DEFAULT_CLASSIFY[type]:
            raise KeyError("error matched")

        if type == "template":
            return self.get_template_circuit(qubits_interval, max_size, max_depth)
        elif type == "random":
            return self.get_random_circuit(classify, qubits_interval, max_size, max_depth)
        elif type == "algorithm":
            return self.get_algorithm_circuit(classify, qubits_interval, max_size, max_depth)
        else:
            return self.get_benchmark_circuit(classify, qubits_interval, max_size, max_depth)



class path(object):
    """ record the path of bfs

    Attribute:
        father_node(int): father in bfs
        CX_tuple(tuple<int, int>): the way father access to son
    """

    def __init__(self, father_node, control, target):
        """ initial method

        father_node(int): the order of father_node
        control(int): the control bit of CX
        target(int): the target bit of CX
        """
        self.father_node = father_node
        self.CX_tuple = (control, target)


def apply_cx(state, control, target, n):
    """ apply cnot gate to the state

    Args:
        state(int): the state represent the matrix
        control(int): the control index for the cx gate
        target(int): the target index for the cx gate
        n(int): number of qubits in the matrix
    Returns:
        int: the new state after applying the gate
    """

    control_col: int = n * control
    target_col: int = n * target

    for i in range(n):
        if state & (1 << (control_col + i)):
            state ^= (1 << (target_col + i))
    return state


def count(nn):
    ans = 1
    for i in range(nn):
        ans *= ((1 << nn) - (1 << i))
    return ans


def generate_layer(n):
    """ generate combination layer for n qubits(n in [2, 5])

    Args:
        n(int): the qubits of layer, in [2, 5]
    Returns:
        list<list<tuple<int, int>>>: the list of layers
    """
    layers = []

    # single layer
    for i in range(n):
        for j in range(n):
            if i != j:
                layers.append([(i, j)])

    # double layer
    for i in range(n):
        for j in range(n):
            for k in range(n):
                for l in range(n):
                    if i != j and i != k and i != l and j != k and j != l and k != l:
                        layers.append([(i, j), (k, l)])

    return layers



import numpy as np
import random

from QuICT.core.circuit.circuit import Circuit
from QuICT.core.gate import BasicGate, UnitaryGate, Unitary, CompositeGate
from QuICT.core.noise import NoiseModel
from QuICT.core.operator import NoiseGate
from QuICT.core.utils import GateType, matrix_product_to_circuit
import QuICT.ops.linalg.cpu_calculator as CPUCalculator
from QuICT.tools.exception.core import TypeError, ValueError
from QuICT.tools.exception.simulation import SimulationMatrixError, SampleBeforeRunError


class DensityMatrixSimulation:
    """ The Density Matrix Simulator

    Args:
        device (str, optional): The device type, one of [CPU, GPU]. Defaults to "CPU".
        precision (str, optional): The precision for the density matrix, one of [single, double]. Defaults to "double".
        accumulated_mode (bool): If True, calculated density matrix with Kraus Operators in NoiseGate.
            if True, p = \\sum Ki p Ki^T.conj(). Default to be False.
            [Important: set accumulated_mode for True if you need sample result, otherwise, using False for
            fast simulate time.]
    """
    def __init__(
        self,
        device: str = "CPU",
        precision: str = "double",
        accumulated_mode: bool = False
    ):
        self._device = device
        self._precision = np.complex128 if precision == "double" else np.complex64
        self._accumulated_mode = accumulated_mode
        self._density_matrix = None

        if device == "CPU":
            self._computer = CPUCalculator
            self._array_helper = np
        elif device == "GPU":
            import QuICT.ops.linalg.gpu_calculator as GPUCalculator
            import cupy as cp

            self._computer = GPUCalculator
            self._array_helper = cp
        else:
            raise ValueError("DensityMatrixSimulation.device", "[CPU, GPU]", device)

    def initial_circuit(self, circuit: Circuit, noise_model: NoiseModel):
        """ Initial the qubits, quantum gates and state vector by given quantum circuit. """
        self._origin_circuit = circuit
        self._circuit = circuit if noise_model is None else noise_model.transpile(circuit, self._accumulated_mode)
        self._noise_model = noise_model
        self._qubits = int(circuit.width())

        if self._precision != circuit._precision:
            circuit.convert_precision()

    def initial_density_matrix(self, qubits: int):
        """ Initial density matrix by given qubits number.

        Args:
            qubits (int): the number of qubits.
        """
        self._density_matrix = self._array_helper.zeros((1 << qubits, 1 << qubits), dtype=self._precision)
        if self._device == "CPU":
            self._density_matrix[0, 0] = self._precision(1)
        else:
            self._density_matrix.put((0, 0), self._precision(1))

    def check_matrix(self, matrix):
        """ Density Matrix Validation. """
        if not np.allclose(matrix.T.conjugate(), matrix):
            raise SimulationMatrixError("The conjugate transpose of density matrix do not equal to itself.")

        if not isinstance(matrix, np.ndarray):
            matrix = matrix.get()

        eigenvalues = np.linalg.eig(matrix)[0]
        for ev in eigenvalues:
            if ev < 0 and not np.isclose(ev, 0, rtol=1e-4):
                raise SimulationMatrixError("The eigenvalues of density matrix should be non-negative")

        if not np.isclose(matrix.trace(), 1, rtol=1e-4):
            raise SimulationMatrixError("The sum of trace of density matrix should be 1.")

        return True

    def run(
        self,
        circuit: Circuit,
        density_matrix: np.ndarray = None,
        noise_model: NoiseModel = None,
        use_previous: bool = False
    ) -> np.ndarray:
        """ Simulating the given circuit through density matrix simulator.

        Args:
            circuit (Circuit): The quantum circuit.
            density_matrix (np.ndarray): The initial density matrix.
            noise_model (NoiseModel, optional): The NoiseModel contains NoiseErrors. Defaults to None.
            use_previous (bool, optional): Using the previous state vector. Defaults to False.

        Returns:
            np.ndarray: the density matrix after simulating
        """
        self.initial_circuit(circuit, noise_model)
        # Initial density matrix
        if density_matrix is not None:
            self.check_matrix(density_matrix)
            self._density_matrix = self._array_helper.array(density_matrix, dtype=self._precision)
        elif (self._density_matrix is None or not use_previous):
            self.initial_density_matrix(self._qubits)

        self._run(self._circuit)

        # Check Readout Error in the NoiseModel
        if noise_model is not None:
            noise_model.apply_readout_error(circuit.qubits)

        return self._density_matrix

    def _run(self, noised_circuit):
        # Start simulator
        cgate = CompositeGate()
        cgate._max_qubit = self._qubits
        for gate in noised_circuit.gates:
            # Store continuous BasicGates into cgate
            if isinstance(gate, BasicGate) and gate.type != GateType.measure:
                gate | cgate
                continue

            if cgate.size() > 0:
                self.apply_gates(cgate)
                cgate.clean()
                cgate._max_qubit = self._qubits

            if gate.type == GateType.measure:
                self.apply_measure(gate.targ)
            elif isinstance(gate, NoiseGate):
                self.apply_noise(gate, self._qubits)
            else:
                raise TypeError("DensityMatrixSimulation.run.circuit", "[BasicGate, NoiseGate]", type(gate))

        if cgate.size() > 0:
            self.apply_gates(cgate)

    def apply_gates(self, cgate: CompositeGate):
        """ Simulating Circuit with BasicGates

        dm = M*dm(M.conj)^T

        Args:
            cgate (CompositeGate): The CompositeGate.
        """
        cgate_matrix = cgate.matrix(self._device)
        self._density_matrix = self._computer.dot(
            self._computer.dot(cgate_matrix, self._density_matrix),
            cgate_matrix.conj().T
        )

    def apply_noise(self, noise_gate: NoiseGate, qubits: int):
        """ Simulating NoiseGate.

        dm = /sum K*dm*(K.conj)^T

        Args:
            noise_gate (NoiseGate): The NoiseGate
            qubits (int): The number of qubits in the circuit.
        """
        gate_args = noise_gate.targs
        noised_matrix = self._array_helper.zeros_like(self._density_matrix)
        for kraus_matrix in noise_gate.noise_matrix:
            umat = matrix_product_to_circuit(kraus_matrix, gate_args, qubits, gpu_output=self._device == "GPU")

            noised_matrix += self._computer.dot(
                self._computer.dot(umat, self._density_matrix),
                umat.conj().T
            )

        self._density_matrix = noised_matrix.copy()

    def apply_measure(self, index: int):
        """ Simulating the MeasureGate.

        Args:
            index (int): The index of measured qubit.
        """
        P0 = self._array_helper.array([[1, 0], [0, 0]], dtype=self._precision)
        mea_0 = matrix_product_to_circuit(P0, index, self._qubits, gpu_output=self._device == "GPU")
        prob_0 = self._array_helper.matmul(mea_0, self._density_matrix).trace()
        _1 = random.random() > prob_0
        if not _1:
            U = self._array_helper.matmul(
                mea_0,
                self._array_helper.eye(1 << self._qubits, dtype=self._precision) / self._array_helper.sqrt(prob_0)
            )
            self._density_matrix = self._computer.dot(self._computer.dot(U, self._density_matrix), U.conj().T)
        else:
            P1 = self._array_helper.array([[0, 0], [0, 1]], dtype=self._precision)
            mea_1 = matrix_product_to_circuit(P1, index, self._qubits, gpu_output=self._device == "GPU")

            U = self._array_helper.matmul(
                mea_1,
                self._array_helper.eye(1 << self._qubits, dtype=self._precision) / self._array_helper.sqrt(1 - prob_0)
            )
            self._density_matrix = self._computer.dot(self._computer.dot(U, self._density_matrix), U.conj().T)

        self._circuit.qubits[index].measured = int(_1)

    def sample(self, shots: int) -> list:
        assert (self._density_matrix is not None), \
            SampleBeforeRunError("DensityMatrixSimulation sample without run any circuit.")
        if self._accumulated_mode or self._noise_model is None:
            original_dm = self._density_matrix.copy()

        state_list = [0] * self._density_matrix.shape[0]
        lastcall_per_qubit = self._circuit.get_lastcall_for_each_qubits()
        measured_idx = [
            i for i in range(self._qubits)
            if lastcall_per_qubit[i] != GateType.measure
        ]

        for _ in range(shots):
            for m_id in measured_idx:
                self.apply_measure(m_id)

            if self._noise_model is not None:
                self._noise_model.apply_readout_error(self._circuit.qubits)

            state_list[int(self._circuit.qubits)] += 1
            if self._accumulated_mode or self._noise_model is None:
                self._density_matrix = original_dm.copy()
            else:
                self.initial_density_matrix(self._qubits)
                noised_circuit = self._noise_model.transpile(self._origin_circuit, self._accumulated_mode) \
                    if self._noise_model is not None else self._circuit

                self._run(noised_circuit)

        return state_list



from QuICT.tools.exception import QuICTException


class TypeError(QuICTException):
    """ Type Error in Core Module, including Layout, Fidelity, Qureg, ... """
    def __init__(self, locate: str, require: str, given: str):
        msg = f"{locate}'s type should be {require}, but given {given}."
        super().__init__(1001, msg)


class ValueError(QuICTException):
    def __init__(self, locate: str, require: str, given: str):
        msg = f"{locate}'s value should be {require}, but given {given}."
        super().__init__(1002, msg)


class IndexExceedError(QuICTException):
    def __init__(self, locate: str, require: str, given: str):
        msg = f"The index of {locate} should be {require}, but given {given}."
        super().__init__(1003, msg)


class QubitMeasureError(QuICTException):
    """ Qubit's Measure Error. """
    def __init__(self, msg: str = None):
        super().__init__(1010, msg)


class CheckPointNoChildError(QuICTException):
    """ The CheckPointChildren not found Error. """
    def __init__(self, msg: str = None):
        super().__init__(1011, msg)



from .. import InstructionSet
from ..transform_rule import xyx_rule

from QuICT.core.gate import *

GoogleSet = InstructionSet(
    GateType.fsim,
    [GateType.sx, GateType.sy, GateType.sw, GateType.rx, GateType.ry]
)
GoogleSet.register_one_qubit_rule(xyx_rule)



from QuICT.simulation.state_vector import CircuitSimulator

from .backend.canonical_QAE import amplitude_estimate as canonical_run
from .backend.max_likely_QAE import amplitude_estimate as max_likely_run
from .backend.FQAE import amplitude_estimate as fast_run
from .backend.canonical_QAE import construct_circuit as canonical_circuit
from .utility import StatePreparationInfo, OracleInfo


class QAE:
    _ALLOWED_MODES = {"canonical", "fast", "max_likely"}
    _RUN_METHOD_OF_MODE = {
        "canonical": canonical_run,
        "fast": fast_run,
        "max_likely": max_likely_run,
    }
    _CIRCUIT_METHOD_OF_MODE = {"canonical": canonical_circuit}

    def __init__(self, mode, eps=0.1, simulator=CircuitSimulator(),) -> None:
        """set up QAE algorithm parameters. mode=="canonical" is not suggested for use\
        due to lower success rate than theoretical expectation

        Args:
            mode (string): _description_
            eps (float, optional): _description_. Defaults to 0.1.
            simulator (_type_, optional): _description_. Defaults to CircuitSimulator().
        """
        if mode not in QAE._ALLOWED_MODES:
            raise ValueError(f"allowed mode are {QAE._ALLOWED_MODES}.")
        self.mode = mode
        self.eps = eps
        self.simulator = simulator

    def run(
        self,
        oracle: OracleInfo,
        state_preparation: StatePreparationInfo = None,
    ):
        if self.mode not in QAE._RUN_METHOD_OF_MODE:
            raise ValueError(f"mode {self.mode} does not support run method.")
        if state_preparation is None:
            state_preparation = StatePreparationInfo(n=oracle.n)
        assert state_preparation.n == oracle.n
        return QAE._RUN_METHOD_OF_MODE[self.mode](
            eps=self.eps,
            oracle=oracle,
            state_preparation=state_preparation,
            simulator=self.simulator,
        )

    def circuit(
        self,
        oracle: OracleInfo,
        state_preparation: StatePreparationInfo = None,
    ):
        if self.mode not in QAE._CIRCUIT_METHOD_OF_MODE:
            raise ValueError(f"mode {self.mode} does not support circuit method.")
        if state_preparation is None:
            state_preparation = StatePreparationInfo(n=oracle.n)
        assert state_preparation.n == oracle.n
        return QAE._CIRCUIT_METHOD_OF_MODE[self.mode](
            eps=self.eps,
            oracle=oracle,
            state_preparation=state_preparation
        )



import matplotlib.pyplot as plt
import numpy as np

from QuICT.algorithm.quantum_algorithm.quantum_walk import Graph, QuantumWalk
from QuICT.simulation.state_vector import ConstantStateVectorSimulator


class QuantumWalkSearch(QuantumWalk):
    """ Search algorithm on a hypercube based on quantum walk and Grover.

    https://arxiv.org/pdf/quant-ph/0210064.pdf
    http://dx.doi.org/10.4236/jqis.2015.51002
    """

    def __init__(self, simulator=ConstantStateVectorSimulator()):
        """ Initialize the simulator circuit of quantum random walk.

        Args:
            simulator (Union[ConstantStateVectorSimulator, CircuitSimulator], optional):
                The simulator for simulating quantum circuit. Defaults to ConstantStateVectorSimulator().
        """
        QuantumWalk.__init__(self, simulator)
        self._search = True

    def _is_unit_hamming_distance(self, x, y):
        """ Calculate the hamming distance of two nodes of the n-cube. """
        return str(bin(x ^ y))[2:].count("1") == 1

    def _get_hypercube_edges(self, position):
        """ Get the edges of the n-cube graph. """
        edge = []
        for i in range(position):
            e = list(np.zeros(self._position_qubits, dtype=np.int64))
            for j in range(position):
                if self._is_unit_hamming_distance(i, j):
                    e[int(np.ceil(np.log2(i ^ j)))] = j
            edge.append(e)
        return edge

    def _set_coins(self, r, a_r, a_nr):
        """ Set the coins for marked and unmarked nodes. """
        n = 2 ** self._action_qubits
        if r is None:
            r = self._action_qubits
        assert 0 <= r < n, "r should in the range of [0, n - 1]."
        assert a_r > a_nr, "a_r should larger than a_nr."

        # set C0 to G
        s_c = np.ones((n, n)) / n
        self._coin_unmarked = np.eye(n) - 2 * s_c

        # set C1
        a = np.sqrt(a_r * a_r + (n - 1) * (a_nr * a_nr))
        x = np.zeros((1, n))
        for i in range(n):
            x[0, i] = a_r / a if i == r else a_nr / a
        self._coin_marked = np.eye(n) - 2 * (x.T @ x)

    def run(
        self,
        index_qubits: int,
        targets: list = None,
        step: int = None,
        r: int = None,
        a_r: float = 1,
        a_nr: float = 0,
        switched_time: int = -1,
        shots: int = 1000,
    ):
        """ Execute the quantum walk search with given number of index qubits.

        Args:
            index_qubits (int): The size of the node register.
            targets (list, optional): The indexes of the target elements.
            step (int, optional): The steps of random walk, a step including a coin operator and a shift operator.
            r (int, optional): The 'direction' of inclination of an asymmetric coin. Defaults to action_qubits (c).
            a_r (float, optional): Parameter of the asymmetry degree of the coin. Defaults to 1.
            a_nr (float, optional): Parameter of the asymmetry degree of the coin. Defaults to 0.
            switched_time (int, optional): The number of steps of each coin operator in the vector.
                Defaults to -1, means not switch coin operator.
            shots (int, optional): The repeated times. Defaults to 1000.

        Returns:
            Union[np.ndarray, List]: The state vector or measured states.
        """

        self._position_qubits = index_qubits  # n
        self._action_qubits = int(np.ceil(np.log2(index_qubits)))  # c
        self._total_qubits = self._position_qubits + self._action_qubits
        position = 1 << index_qubits  # N
        self._step = (
            step
            if step is not None and step > 0
            else int(np.ceil(1.0 / np.sqrt(len(targets) / position))) + 1
        )
        self._step = 6
        self._set_coins(r, a_r, a_nr)
        edges = self._get_hypercube_edges(position)
        self._graph = Graph(position, edges, None, switched_time)

        # Validation graph
        assert self._graph.validation(), "The edge's number should be equal."
        # Validation coin operator
        assert (
            targets is not None and len(targets) > 0
        ), "Should provide target indexes."
        for target in targets:
            assert (
                0 <= target < position
            ), "Target should be within the range of values allowed by the index register. "
        self._targets = targets

        # Build random walk circuit
        self._circuit_construct()

        # Simulate the circuit
        _ = self._simulator.run(self._circuit)

        return self._simulator.sample(shots)



import numpy as np
from typing import Dict, List, Union

from .graph import Graph
from QuICT.core import Circuit
from QuICT.core.gate import *
from QuICT.simulation.state_vector import CircuitSimulator


class QuantumWalk:
    """ The Quantum Random Walk Algorithm. """

    @property
    def step(self):
        return self._step

    @property
    def graph(self):
        return str(self._graph)

    @property
    def circuit(self) -> Circuit:
        """ The quantum circuit of the random walk algorithm, including UnitaryGate. """
        return self._circuit

    def __init__(self, simulator=CircuitSimulator()):
        """ Initialize the simulator circuit of quantum random walk.

        Args:
            simulator (Union[ConstantStateVectorSimulator, CircuitSimulator], optional):
                The simulator for simulating quantum circuit. Defaults to CircuitSimulator().
        """
        self._simulator = simulator
        self._step = None
        self._graph = None
        self._shift_operator = None
        self._coin_operator = None
        self._position_qubits = None
        self._action_qubits = None
        self._total_qubits = None
        self._operator_by_position = False
        self._operator_by_time = False

        # parameters only for quantum walk search
        self._search = False
        self._target = None
        self._coin_marked = None
        self._coin_unmarked = None

    def _coin_operator_validation(self, coin_operator):
        """ Validate the operator. """
        if coin_operator is not None:
            assert self._graph.operator_validation(
                coin_operator
            ), "The coin operator should be an unitary matrix."
            if self._search:
                assert coin_operator.shape[0] == 1 << self._total_qubits, (
                    "The coin oracle should be a unitary matrix"
                    "with side length 2 ** totel_qubits. "
                )
            else:
                self._action_qubits = int(np.ceil(np.log2(coin_operator.shape[0])))
        else:
            self._coin_operator = self._graph.operators
            self._action_qubits = self._graph.action_qubits
            self._operator_by_position = True
            self._operator_by_time = self._graph.switched_time > 0

    def _circuit_construct(self):
        """ Construct random walk circuit. """
        # Build Circuit
        self._circuit = Circuit(self._total_qubits)
        if self._search:
            for idx in range(self._total_qubits):
                H | self.circuit(idx)
        for t in range(self.step):
            self._build_coin_operator(t) | self._circuit
            self._build_shift_operator() | self._circuit

    def _build_coin_operator(self, step: int) -> CompositeGate:
        """ Generator action operator. """
        if self._search:
            search_array = np.zeros((self._graph.position, self._graph.position))
            for target in self._targets:
                search_array[target][target] = 1
            coin_oracle = np.kron(
                np.eye(self._graph.position), self._coin_unmarked
            ) + np.kron(search_array, self._coin_marked - self._coin_unmarked)
            return Unitary(coin_oracle)

        action_qubits = [self._position_qubits + i for i in range(self._action_qubits)]
        if not (self._operator_by_position or self._operator_by_time):
            return Unitary(self._coin_operator) & action_qubits

        action_gate = CompositeGate()
        curr_op_idx = (
            (step // self._graph.switched_time) % len(self._graph.operators[0])
            if self._operator_by_time
            else 0
        )
        for i in range(self._graph.position):
            op = self._graph.operators[i][curr_op_idx]
            x_idx = [
                pidx for pidx in range(self._position_qubits) if not (i & 1 << pidx)
            ]
            for xi in x_idx:
                X | action_gate(xi)

            self._mct_generator(op) | action_gate
            for xi in x_idx:
                X | action_gate(xi)
        action_gate.gate_decomposition()
        return action_gate

    def _mct_generator(self, op: np.ndarray) -> UnitaryGate:
        """ Build multi-control-'op' gate. """
        mct_unitary = np.identity(1 << self._total_qubits, dtype=np.complex128)
        op_shape = op.shape
        mct_unitary[-op_shape[0]:, -op_shape[1]:] = op

        return Unitary(mct_unitary) & list(range(self._total_qubits))

    def _build_shift_operator(self):
        """ Generator shift operator. """
        unitary_matrix = np.zeros(
            (1 << self._total_qubits, 1 << self._total_qubits), dtype=np.complex128
        )
        record_idxes = list(range(1 << self._total_qubits))
        for i in range(self._graph.position):
            curr_idx = (1 << self._action_qubits) * i
            for action_state in range(self._graph.action_space):
                related_idx = self._graph.edges[i][action_state] * (
                    1 << self._action_qubits
                )
                unitary_matrix[related_idx + action_state, curr_idx + action_state] = 1
                if related_idx + action_state in record_idxes:
                    record_idxes.remove(related_idx + action_state)

        if len(record_idxes) > 0:
            unitary_matrix[record_idxes, record_idxes] = 1

        return Unitary(unitary_matrix)

    def run(
        self,
        step: int,
        position: int,
        edges: Union[List, Dict] = None,
        operators: Union[List, Dict] = None,
        coin_operator: np.ndarray = None,
        switched_time: int = -1,
        shots: int = 1000,
    ) -> Union[np.ndarray, List]:
        """ Execute the quantum random walk with given steps, graph and coin operator.

        Args:
            step (int): The steps of random walk, a step including a coin operator and a shift operator.
            position (int): The number of graph's vertex.
            edges (Union[List, Dict], optional): The edges of each vertex. Defaults to None.
            operators (Union[List, Dict], optional): The operators of each vertex. Defaults to None.
            switched_time (int, optional): The number of steps of each coin operator in the vector.
                Defaults to -1, means not switch coin operator.
            coin_operator (np.ndarray, optional): The coin operators, the unitary matrix. Defaults to None.
            shots (int, optional): The repeated times. Defaults to 1000.

        Returns:
            Union[np.ndarray, List]: The state vector or measured states.
        """
        self._step = step
        self._graph = Graph(position, edges, operators, switched_time)
        self._coin_operator = coin_operator

        # Validation graph and coin operator
        assert self._graph.validation(), "The edge's number should be equal."
        self._coin_operator_validation(self._coin_operator)
        self._position_qubits = self._graph.position_qubits
        self._total_qubits = self._position_qubits + self._action_qubits

        # Build random walk circuit
        self._circuit_construct()

        # Simulate the quantum walk's circuit
        _ = self._simulator.run(self._circuit)

        return self._simulator.sample(shots)



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/9/12 12:09
# @Author  : Xiaoquan Xu
# @File    : parametertensor.py

import itertools as it
import numpy as np
from numpy import einsum

from QuICT.chemistry.operator.fermion_operator import FermionOperator


def obi_basis_rotation(obi, R):
    """
    Change the basis of a one-body integrals matrix

    Args:
        obi(Ndarray: n*n): one-body integrals of i^ j
        R(Ndarray: n*n): rotation matrix

    Returns:
        Ndarray: n*n one-body integrals of i^ j after rotation
    """
    return einsum("pi,pq,qj->ij", R.conj(), obi, R)


def tbi_basis_rotation(tbi, R):
    """
    Change the basis of a two-body integrals matrix

    Args:
        tbi(Ndarray: n*n*n*n): two-body integrals of i^ j^ s t
        R(Ndarray: n*n): rotation matrix

    Returns:
        Ndarray: n*n*n*n two-body integrals of i^ j^ s t after rotation
    """
    return einsum("pi,qj,pquv,us,vt->ijst", R.conj(), R.conj(), tbi, R, R)


class ParameterTensor:
    """
    Class for the parameters or Hamiltonian

    Attributes:
        const(complex): For a parameter tensor, it should be set to 1
            for multiplying with a Hamitonian
        obi(n*n ndarray): Represent the coefficients of terms of the form
            a^dagger_i a_j, i.e. (1, 0)
        tbi(n*n*n*n ndarray): Represent the coefficients of terms of the form
            a^dagger_i a^dagger_j a_k a_l, i.e. (1, 1, 0, 0)
    """
    def __init__(self, const, obi, tbi):
        self.const = const
        self.obi = obi
        self.tbi = tbi

    def get_fermion_operator(self):
        """
        Transform a Hamiltonian into a fermion-operator polymonial
        """
        n_qubits = self.obi.shape[0]
        fermion_operator = FermionOperator([], self.const)
        for index in it.product(range(n_qubits), repeat=2):
            fermion_operator += FermionOperator(list(zip(index, (1, 0))), self.obi[index])
        for index in it.product(range(n_qubits), repeat=4):
            fermion_operator += FermionOperator(list(zip(index, (1, 1, 0, 0))), self.tbi[index])
        return fermion_operator

    def expectation(self, other):
        """
        Multiply coefficient matrix with hamitonian
        """
        expectation = self.const * other.const
        expectation += np.sum(self.obi * other.obi)
        expectation += np.sum(self.tbi * other.tbi)
        return expectation


def generate_hamiltonian(const, obi, tbi, EQ_TOLERANCE=1.0E-12):
    """
    Double the dimension of obi and tbi,
    then generate an interaction operator as Hamiltonian
    """
    n = obi.shape[0]
    n_qubits = 2 * n
    new_obi = np.zeros((n_qubits, n_qubits))
    new_tbi = np.zeros((n_qubits, n_qubits, n_qubits, n_qubits))

    for p in range(n):
        for q in range(n):
            new_obi[2 * p, 2 * q] = obi[p, q]
            new_obi[2 * p + 1, 2 * q + 1] = obi[p, q]

    for p in range(n):
        for q in range(n):
            for r in range(n):
                for s in range(n):
                    # Mixed spin
                    new_tbi[2 * p, 2 * q + 1, 2 * r + 1, 2 * s] = tbi[p, q, r, s] / 2.
                    new_tbi[2 * p + 1, 2 * q, 2 * r, 2 * s + 1] = tbi[p, q, r, s] / 2.

                    # Same spin
                    new_tbi[2 * p, 2 * q, 2 * r, 2 * s] = tbi[p, q, r, s] / 2.
                    new_tbi[2 * p + 1, 2 * q + 1, 2 * r + 1, 2 * s + 1] = tbi[p, q, r, s] / 2.

    # Truncate.
    new_obi[np.absolute(new_obi) < EQ_TOLERANCE] = 0.
    new_tbi[np.absolute(new_tbi) < EQ_TOLERANCE] = 0.

    return ParameterTensor(const, new_obi, new_tbi)



from typing import List, Tuple

import networkx as nx
import numpy as np
from QuICT.core import *
from .data_factory import DataFactory


class LayoutInfo:
    def __init__(self, layout: Layout) -> None:
        self.layout = layout
        self._mask = None
        self._graph = None
        self._dist = None
        self._edges = None

    @property
    def topo_graph(self) -> nx.Graph:
        if self._graph is None:
            self._graph = DataFactory.get_topo_graph(self.layout)
        return self._graph

    @property
    def topo_dist(self) -> np.ndarray:
        if self._dist is None:
            self._dist = DataFactory.get_topo_dist(topo_graph=self.topo_graph)
        return self._dist

    @property
    def topo_edges(self) -> List[Tuple[int, int]]:
        if self._edges is None:
            self._edges = DataFactory.get_topo_edges(topo_graph=self.topo_graph)
        return self._edges
